#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import bioio

sys.path.insert(0, os.path.abspath(".."))

# -- General configuration ---------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "numpydoc",
    "myst_parser",
]

myst_enable_extensions = [
    "colon_fence",
    "deflist",
    "linkify",
    "substitution",
]

copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

numpydoc_show_class_members = False

sphinx_tabs_disable_tab_closing = True

autoclass_content = "both"
autodoc_mock_imports = ["bioio_base"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

# The main toctree document.
main_doc = "index"

# General information about the project.
project = "bioio"
copyright = "2022"
author = "Eva Maxfield Brown"

# The version info for the project you"re documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = bioio.__version__
# The full version, including alpha/beta/rc tags.
release = bioio.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"
todo_include_todos = False

# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"
# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {}
html_static_path = ["_static"]

# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "bioiodoc"

# -- Options for LaTeX output ------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (
        main_doc,
        "bioio.tex",
        "bioio Documentation",
        "Eva Maxfield Brown",
        "manual",
    ),
]

# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(main_doc, "bioio", "bioio Documentation", [author], 1)]

# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
# dir menu entry, description, category)
texinfo_documents = [
    (
        main_doc,
        "bioio",
        "bioio Documentation",
        author,
        "bioio",
        "Image reading, metadata management, and image writing for Microscopy images in Python",
    ),
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Configuration for tests! There are a whole list of hooks you can define in this file to
run before, after, or to mutate how tests run. Commonly for most of our work, we use
this file to define top level fixtures that may be needed for tests throughout multiple
test files.

In this case, while we aren't using this fixture in our tests, the prime use case for
something like this would be when we want to preload a file to be used in multiple
tests. File reading can take time, so instead of re-reading the file for each test,
read the file once then use the loaded content.

Docs: https://docs.pytest.org/en/latest/example/simple.html
      https://docs.pytest.org/en/latest/plugins.html#requiring-loading-plugins-in-a-test-module-or-conftest-file
"""
import importlib
import logging
import pathlib
import subprocess
import sys
import typing

import dask.array as da
import numpy as np
import pytest

import bioio

###############################################################################

log = logging.getLogger(__name__)

###############################################################################


@pytest.fixture
def sample_text_file(
    tmp_path: pathlib.Path,
) -> typing.Generator[pathlib.Path, None, None]:
    example_file = tmp_path / "temp-example.txt"
    example_file.write_text("just some example text here")
    yield example_file


def np_random_from_shape(
    shape: typing.Tuple[int, ...], **kwargs: typing.Any
) -> np.ndarray:
    return np.random.randint(255, size=shape, **kwargs)


def da_random_from_shape(
    shape: typing.Tuple[int, ...], **kwargs: typing.Any
) -> da.Array:
    return da.random.randint(255, size=shape, **kwargs)


array_constructor = pytest.mark.parametrize(
    "array_constructor",
    [np_random_from_shape, da_random_from_shape],
)

DUMMY_PLUGIN_NAME = "dummy-plugin"
DUMMY_PLUGIN_PATH = pathlib.Path(__file__).parent / DUMMY_PLUGIN_NAME
ACCEPTING_PLUGIN_NAME = "accepting-plugin"
ACCEPTING_PLUGIN_PATH = pathlib.Path(__file__).parent / ACCEPTING_PLUGIN_NAME


class InstallPackage:
    def __init__(
        self, package_path: typing.Union[str, pathlib.Path], package_name: str
    ):
        if isinstance(package_path, pathlib.Path):
            self.package_path = str(package_path)
        else:
            self.package_path = package_path
        self.package_name = package_name

    def __enter__(self) -> "InstallPackage":
        # Install the plugin
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", self.package_path]
        )
        importlib.reload(bioio)
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[type],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[typing.Type[typing.Any]],
    ) -> None:
        # Uninstall the plugin
        subprocess.check_call(
            [sys.executable, "-m", "pip", "uninstall", "-y", self.package_name]
        )


@pytest.fixture
def dummy_plugin() -> typing.Generator[str, None, None]:
    with InstallPackage(package_path=DUMMY_PLUGIN_PATH, package_name=DUMMY_PLUGIN_NAME):
        yield DUMMY_PLUGIN_NAME


@pytest.fixture
def accepting_plugin() -> typing.Generator[str, None, None]:
    with InstallPackage(
        package_path=ACCEPTING_PLUGIN_PATH, package_name=ACCEPTING_PLUGIN_NAME
    ):
        yield ACCEPTING_PLUGIN_NAME



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import typing

import numpy as np
from ome_types.model.simple_types import PixelType

###############################################################################

log = logging.getLogger(__name__)

###############################################################################


def generate_ome_channel_id(image_id: str, channel_id: typing.Union[str, int]) -> str:
    """
    Naively generates the standard OME channel ID using a provided ID.

    Parameters
    ----------
    image_id: str
        An image id to pull the image specific index from.
        See: `generate_ome_image_id` for more details.
    channel_id: Union[str, int]
        A string or int representing the ID for a channel.
        In the context of the usage of this function, this is usually used with the
        index of the channel.

    Returns
    -------
    ome_channel_id: str
        The OME standard for channel IDs.

    Notes
    -----
    ImageIds are usually: "Image:0", "Image:1", or "Image:N",
    ChannelIds are usually the combination of image index + channel index --
    "Channel:0:0" for the first channel of the first image for example.
    """
    # Remove the prefix 'Image:' to get just the index
    image_index = image_id.replace("Image:", "")
    return f"Channel:{image_index}:{channel_id}"


def generate_ome_image_id(image_id: typing.Union[str, int]) -> str:
    """
    Naively generates the standard OME image ID using a provided ID.

    Parameters
    ----------
    image_id: Union[str, int]
        A string or int representing the ID for an image.
        In the context of the usage of this function, this is usually used with the
        index of the scene / image.

    Returns
    -------
    ome_image_id: str
        The OME standard for image IDs.
    """
    return f"Image:{image_id}"


def dtype_to_ome_type(npdtype: np.dtype) -> PixelType:
    """
    Convert numpy dtype to OME PixelType

    Parameters
    ----------
    npdtype: numpy.dtype
        A numpy datatype.

    Returns
    -------
    ome_type: PixelType
        One of the supported OME Pixels types

    Raises
    ------
    ValueError
        No matching pixel type for provided numpy type.
    """
    ometypedict = {
        np.dtype(np.int8): PixelType.INT8,
        np.dtype(np.int16): PixelType.INT16,
        np.dtype(np.int32): PixelType.INT32,
        np.dtype(np.uint8): PixelType.UINT8,
        np.dtype(np.uint16): PixelType.UINT16,
        np.dtype(np.uint32): PixelType.UINT32,
        np.dtype(np.float32): PixelType.FLOAT,
        np.dtype(np.float64): PixelType.DOUBLE,
        np.dtype(np.complex64): PixelType.COMPLEXFLOAT,
        np.dtype(np.complex128): PixelType.COMPLEXDOUBLE,
    }
    ptype = ometypedict.get(npdtype)
    if ptype is None:
        raise ValueError(f"Ome utils can't resolve pixel type: {npdtype.name}")
    return ptype


def ome_to_numpy_dtype(ome_type: PixelType) -> np.dtype:
    """
    Convert OME PixelType to numpy dtype

    Parameters
    ----------
    ome_type: PixelType
        One of the supported OME Pixels types

    Returns
    -------
    npdtype: numpy.dtype
        A numpy datatype.

    Raises
    ------
    ValueError
        No matching numpy type for the provided pixel type.
    """
    ometypedict: typing.Dict[PixelType, np.dtype] = {
        PixelType.INT8: np.dtype(np.int8),
        PixelType.INT16: np.dtype(np.int16),
        PixelType.INT32: np.dtype(np.int32),
        PixelType.UINT8: np.dtype(np.uint8),
        PixelType.UINT16: np.dtype(np.uint16),
        PixelType.UINT32: np.dtype(np.uint32),
        PixelType.FLOAT: np.dtype(np.float32),
        PixelType.DOUBLE: np.dtype(np.float64),
        PixelType.COMPLEXFLOAT: np.dtype(np.complex64),
        PixelType.COMPLEXDOUBLE: np.dtype(np.complex128),
    }
    nptype = ometypedict.get(ome_type)
    if nptype is None:
        raise ValueError(f"Ome utils can't resolve pixel type: {ome_type.value}")
    return nptype



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
from io import StringIO

import numpy as np

import bioio

from ..plugins import dump_plugins


def test_dump_plugins(dummy_plugin: str) -> None:
    # Capture the output of dump_plugins
    old_stdout = sys.stdout
    sys.stdout = StringIO()
    try:
        # Disable plugin caching to prevent other tests from interfering
        dump_plugins(use_cache=False)
        output = sys.stdout.getvalue()
    finally:
        sys.stdout = old_stdout

    # Check if package name is in the output
    assert dummy_plugin in output


def test_plugin_feasibility_report(dummy_plugin: str) -> None:
    # Arrange
    test_image = np.random.rand(10, 10)
    expected_error_msg = "missing 1 required positional argument: 'path'"

    # Act
    actual_output = bioio.plugin_feasibility_report(test_image)

    # Assert
    assert actual_output["ArrayLike"].supported is True
    assert actual_output["ArrayLike"].error is None
    assert actual_output[dummy_plugin].supported is False
    assert expected_error_msg in (actual_output[dummy_plugin].error or "")



from importlib.metadata import entry_points

import pytest


def test_dummy_writer_discovery_and_api(dummy_plugin: str) -> None:
    # Entry-point registration
    eps = entry_points(group="bioio.writers")
    names = {ep.name for ep in eps}
    assert "DummyWriter" in names, f"found: {sorted(names)}"

    # Public API (__all__)
    import bioio.writers as pkg

    assert "DummyWriter" in getattr(pkg, "__all__", [])

    # Import works (dynamic, so ignore mypy attr-defined error)
    from bioio.writers import DummyWriter  # type: ignore[attr-defined]

    assert DummyWriter.__name__ == "DummyWriter"


def test_dummy_writer_save_stub(dummy_plugin: str) -> None:
    # The save() stub should raise NotImplementedError
    from bioio.writers import DummyWriter  # type: ignore[attr-defined]

    with pytest.raises(NotImplementedError):
        DummyWriter.save(data=[1], uri="unused", dim_order="XYZ")



from typing import Any

import bioio_base as biob
from bioio_base.writer import Writer


class DummyWriter(Writer):
    """
    A dummy writer for testing bioio.writers entry-point discovery.
    """

    @staticmethod
    def save(
        data: biob.types.ArrayLike,
        uri: biob.types.PathLike,
        dim_order: str = biob.dimensions.DEFAULT_DIMENSION_ORDER,
        **kwargs: Any,
    ) -> None:
        """
        Stub implementation that deliberately isn’t implemented.
        """
        raise NotImplementedError("DummyWriter.save is a stub")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pathlib
from importlib.metadata import EntryPoint

import bioio_base as biob
import numpy as np
import pytest
from bioio_base import exceptions
from bioio_base.reader import Reader
from bioio_base.reader_metadata import ReaderMetadata
from bioio_base.types import ImageLike

from bioio import BioImage
from bioio.array_like_reader import ArrayLikeReader


def test_bioimage_with_text_file(sample_text_file: pathlib.Path) -> None:
    with pytest.raises(biob.exceptions.UnsupportedFileFormatError):
        BioImage(sample_text_file)


def test_bioimage_with_missing_file(tmp_path: pathlib.Path) -> None:
    # Construct full filepath
    uri = tmp_path / "does-not-exist-klafjjksdafkjl.bad"
    with pytest.raises(biob.exceptions.UnsupportedFileFormatError):
        BioImage(uri)


def test_bioimage_determine_arraylike() -> None:
    # Arrage
    test_image = np.random.rand(10, 10)

    # Act
    result = BioImage.determine_plugin(test_image)

    # Assert
    assert isinstance(result.entrypoint, EntryPoint)
    assert isinstance(result.metadata, ReaderMetadata)
    assert isinstance(result.timestamp, float)
    assert issubclass(result.metadata.get_reader(), Reader)
    assert result.metadata.get_reader() is ArrayLikeReader


@pytest.mark.parametrize(
    "image, reader_class, expected_exception",
    [
        (
            "s3://my_bucket/my_file.tiff",
            ArrayLikeReader,
            exceptions.UnsupportedFileFormatError,
        ),
        (np.random.rand(10, 10), Reader, exceptions.UnsupportedFileFormatError),
        (
            ["s3://my_bucket/my_file.tiff", "s3://my_bucket/my_file.tiff"],
            ArrayLikeReader,
            exceptions.UnsupportedFileFormatError,
        ),
        (
            [np.random.rand(10, 10), np.random.rand(10, 10)],
            Reader,
            exceptions.UnsupportedFileFormatError,
        ),
    ],
)
def test_bioimage_submission_data_reader_type_alignment(
    image: ImageLike, reader_class: Reader, expected_exception: Exception
) -> None:
    with pytest.raises(expected_exception):
        BioImage(image, reader=reader_class)


def test_bioimage_attempts_s3_read_with_anon_attr(
    sample_text_file: pathlib.Path, dummy_plugin: str
) -> None:
    # Arrange
    from dummy_plugin import Reader as DummyReader

    err_msg = "anon is not True"

    class FakeReader(DummyReader):
        def __init__(self, _: ImageLike, fs_kwargs: dict = {}):
            if fs_kwargs.get("anon", False) is not True:
                raise biob.exceptions.UnsupportedFileFormatError(
                    "test", "test", err_msg
                )

    # Act / Assert
    BioImage("s3://this/could/go/anywhere.ome.zarr", reader=FakeReader)

    # (sanity-check) Make sure it would have failed if not an S3 URI
    # that gets given the anon attribute
    with pytest.raises(biob.exceptions.UnsupportedFileFormatError) as err:
        BioImage(sample_text_file, reader=FakeReader)

    assert err_msg in str(err.value)


def test_bioimage_can_ignore_query_strings(accepting_plugin: str) -> None:
    # The accepting-plugin claims to support .czi files, and it will not error
    # when instantiating a reader. The following should succeed if bioio correctly
    # ignores the query string when evaluating plugin support.
    BioImage(
        "https://allencell.s3.amazonaws.com/aics/hipsc_12x_overview_image_dataset/"
        "stitchedwelloverviewimagepath/05080558_3500003720_10X_20191220_D3.czi"
        "?versionId=_KYMRhRvKxnu727ssMD2_fZD5CmQMNw6"
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import dask.array as da
import numpy as np
import xarray as xr
from bioio_base import constants, exceptions
from bioio_base.dimensions import (
    DEFAULT_DIMENSION_ORDER,
    DEFAULT_DIMENSION_ORDER_WITH_SAMPLES,
    DimensionNames,
)
from bioio_base.reader import Reader
from bioio_base.reader_metadata import ReaderMetadata
from bioio_base.types import MetaArrayLike, PhysicalPixelSizes

from .ome_utils import generate_ome_channel_id, generate_ome_image_id

###############################################################################

log = logging.getLogger(__name__)

###############################################################################


class ArrayLikeReader(Reader):
    """
    A catch all for numpy, dask, or xarray to Reader interface.

    See Notes for more details.

    Parameters
    ----------
    image: Union[List[MetaArrayLike], MetaArrayLike]
        A single, numpy ndarray, dask Array, or xarray DataArray, or list of many.
        If provided a list, each item in the list will be exposed through the scene API.
        If provided an xarray DataArray alone or as an element of the list, the
        dim_order and channel_names kwargs are ignored if there non-xarray
        default dimension (or channel coordinate) information attached the xarray
        object. The channel and dimension information are updated independent of the
        other. If either is using xarray default values, they will be replaced by
        AICSImageIO defaults will be added.

    dim_order: Optional[Union[List[str], str]]
        A string of dimensions to be applied to all array(s) or a
        list of string dimension names to be mapped onto the list of arrays
        provided to image. I.E. "TYX".
        Default: None (guess dimensions for single array or multiple arrays)

    channel_names: Optional[Union[List[str], List[List[str]]]]
        A list of string channel names to be applied to all array(s) or a
        list of lists of string channel names to be mapped onto the list of arrays
        provided to image.
        Default: None (create OME channel IDs for names for single or multiple arrays)

    physical_pixel_sizes: Optional[
        Union[List[float], Dict[str, float], PhysicalPixelSizes]
    ]
        A specification of this image's physical pixel sizes. Can be provided as
        a list, dict or PhysicalPixelSizes object. If a list is passed, the assumed
        order is [Z, Y, X]. If a dict is passed, it must contain "Z", "Y" and "X"
        as keys.
        Default: None

    Raises
    ------
    exceptions.ConflictingArgumentsError
        Raised when the number of scenes provided is different from the number of items
        provided to the metadata parameters if as a list.

    exceptions.ConflictingArgumentsError
        Raised when channel_names is provided but the channel dimension was
        either not guessed or not provided in dim_order.

    ValueError
        Provided dim_order string or channel_names are not the same length as
        the number of dimensions or the size of the channel dimensions for the array at
        the matching index.

    Notes
    -----
    If you want to combine multiple numpy ndarrays or dask arrays with xarray DataArrays
    and attach metadata to all non-xarray.DataArrays, you can do so, the metadata for
    the xarray DataArray scenes will simply be ignored if it the non-xarray defaults.

    In such cases, it is recommended that you provide metadata for those scenes as
    None. I.E.

    >>> some_metadata_attached_xr = ...
    ... some_np = ...
    ... some_dask = ...
    ... reader = ArrayLikeReader(
    ...     image=[some_metadata_attached_xr, some_np, some_dask],
    ...     dim_order=[None, "CTYX", "ZTYX"],
    ...     channel_names=[None, ["A", "B", C"], None],
    ... )

    Will create a three scene ArrayLikeReader with the metadata and coordinate
    information filled xarray DataArray as the first scene, a numpy array with CTYX as
    the dimensions and ["A", "B", "C"] as the channel names, and a dask array with ZTYX
    as the dimensions and no channel names (as there is no channel dimension).
    """

    _scenes: Optional[Tuple[str, ...]] = None

    @staticmethod
    def _is_supported_image(  # type: ignore
        image: Union[List[MetaArrayLike], MetaArrayLike], *args: Any, **kwargs: Any
    ) -> bool:
        if isinstance(image, list):
            values = image
        else:
            values = [image]
        for scene in values:
            if isinstance(scene, (np.ndarray, da.Array, xr.DataArray)):
                continue
            else:
                raise exceptions.UnsupportedFileFormatError(
                    "ArrayLikeReader",
                    f"Unsupported image type: {type(image)}. "
                    + "ArrayLikeReader supported types are numpy ndarray, dask Array,"
                    + "or xarray DataArray.",
                )
        # If we got here, we have a supported image type
        return True

    @staticmethod
    def _guess_dim_order(shape: Tuple[int, ...]) -> str:
        """
        Given an image shape attempts to guess the dimension order.

        Parameters
        ----------
        shape: Tuple[int, ...]
            Tuple of the image array's dimensions.

        Returns
        -------
        dim_order: str
            The guessed dimension order.

        Raises
        ------
        exceptions.InvalidDimensionOrderingError
            Raised when the shape has more than dimensions than the ArrayLikeReader
            has a default order available to support.
        """
        if len(shape) > len(DEFAULT_DIMENSION_ORDER):
            if len(shape) > len(DEFAULT_DIMENSION_ORDER_WITH_SAMPLES):
                raise exceptions.InvalidDimensionOrderingError(
                    "Unable to guess dimension order for array-like images with "
                    + f"more than {len(DEFAULT_DIMENSION_ORDER_WITH_SAMPLES)} "
                    + f"dimensions. Found {len(shape)} dimensions."
                )

            return DEFAULT_DIMENSION_ORDER_WITH_SAMPLES[
                len(DEFAULT_DIMENSION_ORDER_WITH_SAMPLES) - len(shape) :
            ]

        return DEFAULT_DIMENSION_ORDER[len(DEFAULT_DIMENSION_ORDER) - len(shape) :]

    def __init__(
        self,
        image: Union[List[MetaArrayLike], MetaArrayLike],
        dim_order: Optional[Union[List[str], str]] = None,
        channel_names: Optional[Union[List[str], List[List[str]]]] = None,
        physical_pixel_sizes: Optional[
            Union[List[float], Dict[str, float], PhysicalPixelSizes]
        ] = None,
        **kwargs: Any,
    ):
        # Enforce valid image
        self._is_supported_image(image)

        # General note
        # Any time we do a `channel_names[0]` it's because the type check for
        # channel names is a List[List[str]], so by checking the first element we should
        # be getting back a list or a string. Anything else will error.

        # The result of all of this validation and pre-compute work is that at the end
        # of this init, we should have n-number of constructed xarray objects created
        # from the parameters provided that we then just pass-through to for all other
        # standard Reader operations

        # If metadata is attached as lists
        # Enforcing matching shape
        if isinstance(image, list):
            if isinstance(dim_order, list):
                # Check dim order
                if len(dim_order) != len(image):
                    raise exceptions.ConflictingArgumentsError(
                        f"ArrayLikeReader received a list of arrays to use as scenes "
                        f"but the provided list of dimension order strings is of "
                        f"different length. "
                        f"Number of provided scenes: {len(image)}, "
                        f"Number of provided dimension order strings: "
                        f"{len(dim_order)}"
                    )

            # Check channel names
            if channel_names is not None:
                if isinstance(channel_names[0], list):
                    if len(channel_names) != len(image):
                        raise exceptions.ConflictingArgumentsError(
                            f"ArrayLikeReader received a list of arrays to use as "
                            f"scenes but the provided list of channel names is "
                            f"of different length. "
                            f"Number of provided scenes: {len(image)}, "
                            f"Number of provided channel names: "
                            f"{len(channel_names)}"
                        )

        # If metadata is attached as singles
        # but many scenes provided, expand
        if isinstance(image, list):
            if dim_order is None or isinstance(dim_order, str):
                dim_order = [dim_order for i in range(len(image))]  # type: ignore
            if channel_names is None or isinstance(channel_names[0], str):
                channel_names = [  # type: ignore
                    channel_names for i in range(len(image))
                ]

        # Set all kwargs to lists for standard interface
        if not isinstance(image, list):
            image = [image]
        if not isinstance(dim_order, list):
            dim_order = [dim_order]  # type: ignore
        if channel_names is None:
            channel_names = [channel_names]  # type: ignore
        # Also wrap the channel names list if they were provided
        # but only a single scene was
        elif len(image) == 1 and not isinstance(channel_names[0], list):
            channel_names = [channel_names]  # type: ignore

        # Store image(s)
        self._all_scenes = image

        # Validate and store dims
        self._scene_dims_list = []
        for i, dims_string in enumerate(dim_order):
            this_scene = self._all_scenes[i]
            # Provided None, guess
            if dims_string is None:
                # Get dims from xarray
                if isinstance(this_scene, xr.DataArray):
                    # Guess for xarray when they guessed
                    # http://xarray.pydata.org/en/stable/data-structures.html?highlight=dim_n#creating-a-dataarray
                    # See "dim_N"
                    if this_scene.dims[0] == "dim_0":
                        log.debug(
                            "Overrode the default xarray dimensions for AICSImage "
                            "guessed dimensions."
                        )
                        # Rename the dimensions from "dim_N" to just the guess dim
                        # Update scene list in place
                        self._all_scenes[i] = this_scene.rename(
                            {
                                f"dim_{d_index}": d
                                for d_index, d in enumerate(
                                    self._guess_dim_order(this_scene.shape)
                                )
                            }
                        )

                # Guess for non xarray data
                self._scene_dims_list.append(self._guess_dim_order(this_scene.shape))

            # Provided Some, validate
            else:
                if len(dims_string) == len(this_scene.shape):
                    # Update dims for xarray
                    if isinstance(this_scene, xr.DataArray):
                        # Guess for xarray when they guessed
                        # http://xarray.pydata.org/en/stable/data-structures.html?highlight=dim_n#creating-a-dataarray
                        # See "dim_N"
                        if this_scene.dims[0] == "dim_0":
                            log.debug(
                                "Overrode the default xarray dimensions for AICSImage "
                                "provided dimensions."
                            )
                            # Rename the dimensions from "dim_N" to just the guess dim
                            # Update scene list in place
                            self._all_scenes[i] = this_scene.rename(
                                {
                                    f"dim_{d_index}": d
                                    for d_index, d in enumerate(dims_string)
                                }
                            )

                    self._scene_dims_list.append(dims_string)
                else:
                    raise ValueError(
                        f"Provided dimension string does not have the same amount of "
                        f"dimensions as the matching provided array. "
                        f"Provided array shape: {this_scene.shape}, "
                        f"Provided dimension string: {dims_string}"
                    )

        # Validate and store channel_names
        self._scene_channel_names = []
        for s_index, this_scene_channel_names in enumerate(
            channel_names  # type: ignore
        ):
            this_scene = self._all_scenes[s_index]
            this_scene_dims = self._scene_dims_list[s_index]

            # Create channel names is needed
            if DimensionNames.Channel in this_scene_dims:
                channel_dim_index = this_scene_dims.index(DimensionNames.Channel)

                # Handle xarray missing channel names
                if isinstance(this_scene, xr.DataArray):
                    if DimensionNames.Channel not in this_scene.coords:
                        # Use provided
                        if this_scene_channel_names is not None:
                            this_scene.coords[
                                DimensionNames.Channel
                            ] = this_scene_channel_names

                        # Generate
                        else:
                            set_channel_names = []
                            for c_index in range(this_scene.shape[channel_dim_index]):
                                image_id = generate_ome_image_id(s_index)
                                set_channel_names.append(
                                    generate_ome_channel_id(
                                        image_id=image_id, channel_id=c_index
                                    )
                                )

                            this_scene.coords[
                                DimensionNames.Channel
                            ] = set_channel_names

                # Provided None, generate
                if this_scene_channel_names is None:
                    this_scene_channels = []
                    for c_index in range(this_scene.shape[channel_dim_index]):
                        image_id = generate_ome_image_id(s_index)
                        this_scene_channels.append(
                            generate_ome_channel_id(
                                image_id=image_id, channel_id=c_index
                            )
                        )

                    self._scene_channel_names.append(this_scene_channels)

                # Provided Some, validate
                else:
                    if (
                        len(this_scene_channel_names)
                        != this_scene.shape[channel_dim_index]
                    ):
                        raise ValueError(
                            f"Provided channel names list does not match the size of "
                            f"channel dimension for the provided array. "
                            f"Provided array shape: {this_scene.shape}, "
                            f"Channel dimension size: "
                            f"{this_scene.shape[channel_dim_index]}, "
                            f"Provided channel names: {this_scene_channel_names}"
                        )
                    else:
                        self._scene_channel_names.append(this_scene_channel_names)

            # Raise error when channel names were provided when they shouldn't have been
            else:
                if this_scene_channel_names is not None:
                    raise ValueError(
                        f"Received channel names for array without channel dimension. "
                        f"Provided array shape: {this_scene.shape}, "
                        f"Provided (or guessed) dimensions: {this_scene_dims}, "
                        f"Provided channel names: {this_scene_channel_names}"
                    )
                else:
                    self._scene_channel_names.append(this_scene_channel_names)

        # Construct full xarrays
        # All data arrays in this list are dask Arrays
        self._xr_darrays = []
        for scene_data, dims, this_scene_channel_names in zip(
            self._all_scenes, self._scene_dims_list, self._scene_channel_names
        ):
            # Handle simple case of provided a DataArray
            if isinstance(scene_data, xr.DataArray):
                # Set metadata to point at existing attrs
                scene_data.attrs[constants.METADATA_UNPROCESSED] = scene_data.attrs

                # If the data backing the xarray is dask
                # just append
                if isinstance(scene_data.data, da.Array):
                    self._xr_darrays.append(scene_data)

                # If the data backing the xarray is numpy
                # copy the array but use dask from numpy
                else:
                    self._xr_darrays.append(
                        scene_data.copy(data=da.from_array(scene_data.data))
                    )

            # Handle non-xarray cases
            else:
                dims_list = list(dims)
                coords = {}
                if DimensionNames.Channel in dims_list:
                    coords[DimensionNames.Channel] = this_scene_channel_names

                # Handle dask
                if isinstance(scene_data, np.ndarray):
                    scene_data = da.from_array(scene_data)

                # Append the dask backed array
                self._xr_darrays.append(
                    xr.DataArray(
                        data=scene_data,
                        dims=dims_list,
                        coords=coords,
                        attrs={constants.METADATA_UNPROCESSED: None},
                    )
                )

        if isinstance(physical_pixel_sizes, PhysicalPixelSizes):
            self._physical_pixel_sizes = physical_pixel_sizes
        elif isinstance(physical_pixel_sizes, (list, tuple)):
            self._physical_pixel_sizes = PhysicalPixelSizes(*physical_pixel_sizes)
        elif isinstance(physical_pixel_sizes, dict):
            self._physical_pixel_sizes = PhysicalPixelSizes(**physical_pixel_sizes)
        else:
            self._physical_pixel_sizes = None

    @property
    def scenes(self) -> Tuple[str, ...]:
        if self._scenes is None:
            self._scenes = tuple(
                generate_ome_image_id(i) for i in range(len(self._all_scenes))
            )

        return self._scenes

    def _read_delayed(self) -> xr.DataArray:
        return self._xr_darrays[self.current_scene_index]

    def _read_immediate(self) -> xr.DataArray:
        return self._xr_darrays[self.current_scene_index].copy(
            data=self._xr_darrays[self.current_scene_index].data.compute()
        )

    @property
    def physical_pixel_sizes(self) -> PhysicalPixelSizes:
        """
        Returns
        -------
        sizes: PhysicalPixelSizes
            Using available metadata, the floats representing physical pixel sizes for
            dimensions Z, Y, and X.

        Notes
        -----
        We currently do not handle unit attachment to these values. Please see the file
        metadata for unit information.
        """
        return self._physical_pixel_sizes


class ArrayLikeReaderMetadata(ReaderMetadata):
    @staticmethod
    def get_supported_extensions() -> List[str]:
        # Since ArrayLikeReader supports numpy, dask, and xarray objects it does
        # not have file extensions
        return []

    @staticmethod
    def get_reader() -> Reader:
        return ArrayLikeReader



#!/usr/bin/env python

"""Top-level package for bioio."""

from importlib.metadata import PackageNotFoundError, version

try:
    from bioio_base.dimensions import DimensionNames, Dimensions
    from bioio_base.standard_metadata import StandardMetadata
    from bioio_base.types import (
        ArrayLike,
        DimensionProperties,
        ImageLike,
        MetaArrayLike,
        PathLike,
        PhysicalPixelSizes,
        Scale,
        TimeInterval,
    )
    from bioio_base.writer import Writer
except ImportError:
    # Safe fallback during install/docs build
    _fallback_symbols = [
        "ArrayLike",
        "DimensionNames",
        "DimensionProperties",
        "Dimensions",
        "ImageLike",
        "MetaArrayLike",
        "PathLike",
        "PhysicalPixelSizes",
        "Scale",
        "StandardMetadata",
        "TimeInterval",
        "Writer",
    ]
    globals().update({name: None for name in _fallback_symbols})


from .bio_image import BioImage
from .plugins import plugin_feasibility_report

try:
    __version__ = version("bioio")
except PackageNotFoundError:
    __version__ = "uninstalled"

__author__ = "Eva Maxfield Brown, Dan Toloudis, BioIO Contributors"
__email__ = "evamaxfieldbrown@gmail.com, danielt@alleninstitute.org"

__all__ = [
    "ArrayLike",
    "BioImage",
    "DimensionNames",
    "DimensionProperties",
    "Dimensions",
    "ImageLike",
    "MetaArrayLike",
    "PathLike",
    "PhysicalPixelSizes",
    "Scale",
    "StandardMetadata",
    "TimeInterval",
    "Writer",
    "plugin_feasibility_report",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import re
import sys
from collections import OrderedDict
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, get_args

import semver

if sys.version_info < (3, 10):
    from importlib_metadata import EntryPoint, entry_points, requires
else:
    from importlib.metadata import entry_points, EntryPoint, requires

import time
from typing import Dict, List, NamedTuple, Optional, Tuple

from bioio_base.reader import Reader
from bioio_base.reader_metadata import ReaderMetadata
from bioio_base.types import ArrayLike, ImageLike, PathLike

from .array_like_reader import ArrayLikeReader, ArrayLikeReaderMetadata

###############################################################################

log = logging.getLogger(__name__)

###############################################################################

BIOIO_DIST_NAME = "bioio"
BIOIO_BASE_DIST_NAME = "bioio-base"


class PluginEntry(NamedTuple):
    entrypoint: EntryPoint
    metadata: ReaderMetadata
    timestamp: float


# global cache of plugins
plugins_by_ext_cache: OrderedDict[str, List[PluginEntry]] = OrderedDict()


def check_type(image: ImageLike, reader_class: Reader) -> bool:
    """
    Check if the provided image is compatible with the specified reader class.

    Parameters
    ----------
    image : ImageLike
        The image to be checked. It can be a PathLike, ArrayLike, MetaArrayLike,
        a list of MetaArrayLike, or a list of PathLike.

    reader_class : Reader
        The reader class to be checked against.
    Returns
    -------
    bool
        Returns True if the image is compatible with the reader class, False otherwise.
    """
    arraylike_types = tuple(get_args(ArrayLike))
    pathlike_types = tuple(get_args(PathLike))

    # Check if image is type ArrayLike or list of ArrayLike and reader
    # is not an ArrayLikeReader
    if (
        isinstance(image, arraylike_types)
        or (
            isinstance(image, list)
            and all(isinstance(item, arraylike_types) for item in image)
        )
    ) and not (reader_class is ArrayLikeReader):
        return False

    # Check if image is type PathLike or list of PathLike and reader
    # is an ArrayLikeReader
    if (
        isinstance(image, pathlike_types)
        or (
            isinstance(image, list)
            and all(isinstance(item, pathlike_types) for item in image)
        )
    ) and (reader_class is ArrayLikeReader):
        return False

    return True


def get_array_like_plugin() -> PluginEntry:
    """
    Create and return a PluginEntry for ArrayLikeReader.
    """
    entrypoint = EntryPoint(
        name="ArrayLikeReader",
        group="readers",
        value=".array_like_reader.ArrayLikeReader",
    )
    metadata = ArrayLikeReaderMetadata()
    timestamp = time.time()
    return PluginEntry(entrypoint=entrypoint, metadata=metadata, timestamp=timestamp)


def insert_sorted_by_timestamp(list: List[PluginEntry], item: PluginEntry) -> None:
    """
    Insert into list of PluginEntrys sorted by their timestamps (install dates)
    """
    for i, other in enumerate(list):
        if item.timestamp > other.timestamp:
            list.insert(i, item)
            return
    list.append(item)


def get_dependency_version_range_for_distribution(
    distribution_name: str, dependency_name: str
) -> Tuple[Optional[str], Optional[str]]:
    """
    Retrieves the minimum and maximum versions of the dependency as specified
    by the given distribution.
    `None` as the minimum or maximum represents no minimum or maximum (respectively)
    """
    all_distribution_deps = requires(distribution_name) or []
    matching_distribution_deps = [
        dep for dep in all_distribution_deps if dep.startswith(dependency_name)
    ]
    if len(matching_distribution_deps) != 1:
        raise ValueError(
            f"Expected to find 1 `{dependency_name}` dependency for "
            f"`{distribution_name}`, instead found {len(matching_distribution_deps)}"
        )

    # Versions from importlib are formatted like "<dependency name> (<0.19,>=0.18)"
    # so this regex needs to find the inner part of the parentheses to get the
    # version specifications
    version_specification_match = re.findall(r"\(.*?\)", matching_distribution_deps[0])

    minimum_dependency_version = None
    maximum_dependency_version = None
    if version_specification_match:
        # A distribution can specify a minimum and maximum (or force an exact) version
        # so multiple versions need to be parsed here
        version_specifications = version_specification_match[0].split(",")
        for version_specification in version_specifications:
            version = "".join(
                [
                    char
                    for char in version_specification
                    if char.isnumeric() or char == "."
                ]
            )

            # Unfortunately not all versions are specified in full SemVer format
            # and instead can be like "0.19" and the "semver" package does not
            # handle these well so we can just fill in the missing .0
            # to make 0.19 -> 0.19.0
            while version.count(".") < 2:
                version += ".0"

            if "<" in version_specification or "==" in version_specification:
                maximum_dependency_version = version

            if ">" in version_specification or "==" in version_specification:
                minimum_dependency_version = version

    return minimum_dependency_version, maximum_dependency_version


def get_plugins(use_cache: bool) -> Dict[str, List[PluginEntry]]:
    """
    Gather a mapping from support file extensions
    to plugins installed that support said extension
    """
    if use_cache and plugins_by_ext_cache:
        return plugins_by_ext_cache.copy()

    plugins = entry_points(group="bioio.readers")

    # Mapping of extensions -> applicable plugins
    # note there can be multiple readers for the same extension
    plugins_by_ext: Dict[str, List[PluginEntry]] = {}

    (
        min_compatible_bioio_base_version,
        _,
    ) = get_dependency_version_range_for_distribution(
        BIOIO_DIST_NAME, BIOIO_BASE_DIST_NAME
    )
    for plugin in plugins:
        (
            _,
            max_bioio_base_version_for_plugin,
        ) = get_dependency_version_range_for_distribution(
            plugin.name, BIOIO_BASE_DIST_NAME
        )
        if (
            min_compatible_bioio_base_version is not None
            and max_bioio_base_version_for_plugin is not None
            and semver.compare(
                max_bioio_base_version_for_plugin,
                min_compatible_bioio_base_version,
            )
            < 0
        ):
            print(
                f"Plugin `{plugin.name}` does not meet "
                f"minimum `{BIOIO_BASE_DIST_NAME}` of "
                f"{min_compatible_bioio_base_version}, ignoring"
            )
        else:
            # ReaderMetadata knows how to instantiate the actual Reader
            reader_meta = plugin.load().ReaderMetadata
            if plugin.dist is not None:
                files = plugin.dist.files
                if files is not None:
                    firstfile = files[0]
                    timestamp = os.path.getmtime(Path(firstfile.locate()).parent)
                else:
                    print(f"No files found for plugin: '{plugin}'")
            else:
                print(f"Could not find distribution for plugin: '{plugin}'")
                timestamp = 0.0

            # Add plugin entry
            plugin_entry = PluginEntry(plugin, reader_meta, timestamp)
            for ext in plugin_entry.metadata.get_supported_extensions():
                ext = ext.lower()
                if not ext.startswith("."):
                    ext = "." + ext

                # Start a new list of plugins for ext if it doesn't exist
                if ext not in plugins_by_ext:
                    plugins_by_ext[ext] = [plugin_entry]
                    continue

                # insert in sorted order (sorted by most recently installed)
                pluginlist = plugins_by_ext[ext]
                insert_sorted_by_timestamp(pluginlist, plugin_entry)

    # Dictionary values (the lists of plugin entries) have already been sorted
    # by timestamp due to the "insert_sorted_by_timestamp" function
    # However, we further want to sort the dictionary keys (the extensions)
    # by length so that longer extensions are checked first.
    # We do not change the order of the plugins within each list.
    plugins_by_ext = OrderedDict(
        sorted(
            # Get the key (suffix) and the value (list of plugins)
            plugins_by_ext.items(),
            # Sort by length of the key (suffix)
            key=lambda ext_and_plugins: len(ext_and_plugins[0]),
            # Reverse so that longer extensions are stored first
            reverse=True,
        )
    )

    # Save copy of plugins to cache then return
    plugins_by_ext_cache.clear()
    plugins_by_ext_cache.update(plugins_by_ext)

    return plugins_by_ext


def dump_plugins(use_cache: bool = True) -> None:
    """
    Report information about plugins currently installed

    Parameters
    ----------
    use_cache : bool
        Whether to use the cached plugins list. Mainly exposed for testing purposes.
    """
    plugins_by_ext = get_plugins(use_cache=use_cache)
    plugin_set = set()
    for _, plugins in plugins_by_ext.items():
        plugin_set.update(plugins)

    for plugin in plugin_set:
        ep = plugin.entrypoint
        print(ep.name)

        # Unpack dist
        dist = ep.dist
        if dist is not None:
            author = dist.metadata.json.get("author")
            license = dist.metadata.json.get("license")

            print(f'  Author  : {author if author is not None else "Not Specified"}')
            print(f"  Version : {dist.version}")
            print(f'  License : {license if license is not None else "Not Specified"}')

            # Unpack files
            files = dist.files
            if files is not None:
                firstfile = files[0]
                t = datetime.fromtimestamp(
                    os.path.getmtime(Path(firstfile.locate()).parent)
                )
                print(f"  Date    : {t}")
        else:
            print("  No Distribution Found...")

            # print(f"  Description : {ep.dist.metadata['description']}")
        reader_meta = plugin.metadata
        exts = ", ".join(reader_meta.get_supported_extensions())
        print(f"  Supported Extensions : {exts}")
    print("Plugins for extensions:")
    sorted_exts = sorted(plugins_by_ext.keys())
    for ext in sorted_exts:
        plugins = plugins_by_ext[ext]
        print(f"{ext}: {plugins}")


@dataclass
class PluginSupport:
    """Dataclass for reporting a plugins support for a specific image."""

    supported: bool
    error: Optional[str]


def _check_plugin_support(
    plugin: PluginEntry, image: ImageLike, fs_kwargs: Dict[str, Any] = {}
) -> PluginSupport:
    """Helper function to check if a plugin supports the given image."""
    try:
        ReaderClass = plugin.metadata.get_reader()
        supported = ReaderClass.is_supported_image(image=image, fs_kwargs=fs_kwargs)
        return PluginSupport(supported=supported, error=None)
    except Exception as e:
        return PluginSupport(supported=False, error=str(e))


def plugin_feasibility_report(
    image: ImageLike,
    fs_kwargs: Dict[str, Any] = {},
    use_plugin_cache: bool = False,
    **kwargs: Any,
) -> Dict[str, PluginSupport]:
    """
    Generate a feasibility report for each plugin,
    determining if it can handle the specified image.
    """
    plugins_by_ext = get_plugins(use_cache=use_plugin_cache)
    feasibility_report = {}

    # Check each plugin for support
    for plugins in plugins_by_ext.values():
        for plugin in plugins:
            plugin_name = plugin.entrypoint.name
            feasibility_report[plugin_name] = _check_plugin_support(
                plugin, image, fs_kwargs
            )

    # Additional check for ArrayLike support
    try:
        supported = ArrayLikeReader.is_supported_image(image=image, fs_kwargs=fs_kwargs)
        feasibility_report["ArrayLike"] = PluginSupport(supported=supported, error=None)
    except Exception as e:
        feasibility_report["ArrayLike"] = PluginSupport(supported=False, error=str(e))

    # Log feasibility report in a readable format
    log.info("Feasibility Report Summary:")
    for name, status in feasibility_report.items():
        if status.error is not None:
            log.info(f"{name}: Unsupported - Error: {status.error}")
        else:
            log.info(f"{name}: {'Supported' if status.supported else 'Unsupported'}")

    return feasibility_report



from typing import Any, Optional, Tuple

import xarray as xr
from bioio_base.dimensions import Dimensions
from bioio_base.reader import Reader as BaseReader
from fsspec.spec import AbstractFileSystem


class Reader(BaseReader):
    """
    Dummy reader plugin that pretends to read anything
    """

    _xarray_dask_data: Optional["xr.DataArray"] = None
    _xarray_data: Optional["xr.DataArray"] = None
    _mosaic_xarray_dask_data: Optional["xr.DataArray"] = None
    _mosaic_xarray_data: Optional["xr.DataArray"] = None
    _dims: Optional[Dimensions] = None
    _metadata: Optional[Any] = None
    _scenes: Optional[Tuple[str, ...]] = None
    _current_scene_index: int = 0
    _fs: "AbstractFileSystem"
    _path: str

    def __init__(*args: Any, **kwargs: Any):
        pass

    @staticmethod
    def _is_supported_image(fs: "AbstractFileSystem", path: str, **kwargs: Any) -> bool:
        return True

    @property
    def scenes(self) -> Tuple[str, ...]:
        return ()

    def _read_delayed(self) -> "xr.DataArray":
        return xr.DataArray([])

    def _read_immediate(self) -> "xr.DataArray":
        return xr.DataArray([])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import List, Optional, Tuple, Union

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from bioio_base import dimensions, exceptions, test_utilities, types

from ..array_like_reader import ArrayLikeReader
from ..bio_image import BioImage


@pytest.mark.parametrize(
    "image, "
    "dim_order, "
    "channel_names, "
    "set_scene, "
    "expected_scenes, "
    "expected_shape, "
    "expected_dims, "
    "expected_channel_names",
    [
        # Check no metadata provided 2D
        # these are really simple just None checks
        (np.random.rand(1, 1), None, None, "Image:0", ("Image:0",), (1, 1), "YX", None),
        (
            da.random.random((1, 1)),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1),
            "YX",
            None,
        ),
        (
            xr.DataArray(np.random.rand(1, 1)),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1),
            "YX",
            None,
        ),
        (
            xr.DataArray(da.random.random((1, 1))),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1),
            "YX",
            None,
        ),
        # Check no metadata provided 3D
        # these are really simply just None checks
        (
            np.random.rand(1, 1, 1),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        (
            da.random.random((1, 1, 1)),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        (
            xr.DataArray(np.random.rand(1, 1, 1)),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        (
            xr.DataArray(da.random.random((1, 1, 1))),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        # Check no metadata provided 4D
        # these check that channel names are created for all
        # and specifically for xr that channel names are overwritten
        (
            np.random.rand(1, 1, 1, 1),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1),
            "CZYX",
            ["Channel:0:0"],
        ),
        (
            da.random.random((1, 1, 1, 1)),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1),
            "CZYX",
            ["Channel:0:0"],
        ),
        (
            xr.DataArray(np.random.rand(1, 1, 1, 1)),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1),
            "CZYX",
            ["Channel:0:0"],
        ),
        (
            xr.DataArray(da.random.random((1, 1, 1, 1))),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1),
            "CZYX",
            ["Channel:0:0"],
        ),
        # Check no channel names provided 3D but dim_order forces channel dim
        # these check that channel names are created for all
        # and specifically for xr that channel names are overwritten
        (
            np.random.rand(1, 1, 1),
            "CYX",
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "CYX",
            ["Channel:0:0"],
        ),
        (
            da.random.random((1, 1, 1)),
            "CYX",
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "CYX",
            ["Channel:0:0"],
        ),
        (
            xr.DataArray(np.random.rand(1, 1, 1)),
            "CYX",
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "CYX",
            ["Channel:0:0"],
        ),
        (
            xr.DataArray(da.random.random((1, 1, 1))),
            "CYX",
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "CYX",
            ["Channel:0:0"],
        ),
        # Test many scene, same dim_order, first scene
        (
            [np.random.rand(1, 1, 1), np.random.rand(2, 2, 2)],
            "CYX",
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "CYX",
            ["Channel:0:0"],
        ),
        (
            [da.random.random((1, 1, 1)), da.random.random((2, 2, 2))],
            "CYX",
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "CYX",
            ["Channel:0:0"],
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1)),
                xr.DataArray(np.random.rand(2, 2, 2)),
            ],
            "CYX",
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "CYX",
            ["Channel:0:0"],
        ),
        (
            [
                xr.DataArray(da.random.random((1, 1, 1))),
                xr.DataArray(da.random.random((2, 2, 2))),
            ],
            "CYX",
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "CYX",
            ["Channel:0:0"],
        ),
        # Test many scene, same dim_order, second scene
        (
            [np.random.rand(1, 1, 1), np.random.rand(2, 2, 2)],
            "CYX",
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["Channel:1:0", "Channel:1:1"],
        ),
        (
            [da.random.random((1, 1, 1)), da.random.random((2, 2, 2))],
            "CYX",
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["Channel:1:0", "Channel:1:1"],
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1)),
                xr.DataArray(np.random.rand(2, 2, 2)),
            ],
            "CYX",
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["Channel:1:0", "Channel:1:1"],
        ),
        (
            [
                xr.DataArray(da.random.random((1, 1, 1))),
                xr.DataArray(da.random.random((2, 2, 2))),
            ],
            "CYX",
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["Channel:1:0", "Channel:1:1"],
        ),
        # Test many scene, same channel_names, first scene
        (
            [np.random.rand(2, 1, 1, 1), np.random.rand(2, 2, 2, 2)],
            None,
            ["A", "B"],
            "Image:0",
            ("Image:0", "Image:1"),
            (2, 1, 1, 1),
            "CZYX",
            ["A", "B"],
        ),
        (
            [da.random.random((2, 1, 1, 1)), da.random.random((2, 2, 2, 2))],
            None,
            ["A", "B"],
            "Image:0",
            ("Image:0", "Image:1"),
            (2, 1, 1, 1),
            "CZYX",
            ["A", "B"],
        ),
        (
            [
                xr.DataArray(np.random.rand(2, 1, 1, 1)),
                xr.DataArray(np.random.rand(2, 2, 2, 2)),
            ],
            None,
            ["A", "B"],
            "Image:0",
            ("Image:0", "Image:1"),
            (2, 1, 1, 1),
            "CZYX",
            ["A", "B"],
        ),
        (
            [
                xr.DataArray(da.random.random((2, 1, 1, 1))),
                xr.DataArray(da.random.random((2, 2, 2, 2))),
            ],
            None,
            ["A", "B"],
            "Image:0",
            ("Image:0", "Image:1"),
            (2, 1, 1, 1),
            "CZYX",
            ["A", "B"],
        ),
        # Test many scene, same channel_names, second scene
        (
            [np.random.rand(2, 1, 1, 1), np.random.rand(2, 2, 2, 2)],
            None,
            ["A", "B"],
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2, 2),
            "CZYX",
            ["A", "B"],
        ),
        (
            [da.random.random((2, 1, 1, 1)), da.random.random((2, 2, 2, 2))],
            None,
            ["A", "B"],
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2, 2),
            "CZYX",
            ["A", "B"],
        ),
        (
            [
                xr.DataArray(np.random.rand(2, 1, 1, 1)),
                xr.DataArray(np.random.rand(2, 2, 2, 2)),
            ],
            None,
            ["A", "B"],
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2, 2),
            "CZYX",
            ["A", "B"],
        ),
        (
            [
                xr.DataArray(da.random.random((2, 1, 1, 1))),
                xr.DataArray(da.random.random((2, 2, 2, 2))),
            ],
            None,
            ["A", "B"],
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2, 2),
            "CZYX",
            ["A", "B"],
        ),
        # Test many scene, different dim_order, first scene
        (
            [np.random.rand(1, 1, 1), np.random.rand(2, 2, 2)],
            [None, "CYX"],
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        (
            [da.random.random((1, 1, 1)), da.random.random((2, 2, 2))],
            [None, "CYX"],
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1)),
                xr.DataArray(np.random.rand(2, 2, 2)),
            ],
            [None, "CYX"],
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        (
            [
                xr.DataArray(da.random.random((1, 1, 1))),
                xr.DataArray(da.random.random((2, 2, 2))),
            ],
            [None, "CYX"],
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        # Test many scene, different dim_order, second scene
        (
            [np.random.rand(1, 1, 1), np.random.rand(2, 2, 2)],
            [None, "CYX"],
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["Channel:1:0", "Channel:1:1"],
        ),
        (
            [da.random.random((1, 1, 1)), da.random.random((2, 2, 2))],
            [None, "CYX"],
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["Channel:1:0", "Channel:1:1"],
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1)),
                xr.DataArray(np.random.rand(2, 2, 2)),
            ],
            [None, "CYX"],
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["Channel:1:0", "Channel:1:1"],
        ),
        (
            [
                xr.DataArray(da.random.random((1, 1, 1))),
                xr.DataArray(da.random.random((2, 2, 2))),
            ],
            [None, "CYX"],
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["Channel:1:0", "Channel:1:1"],
        ),
        # Test many scene, different dim_order, different channel_names, first scene
        (
            [np.random.rand(1, 1, 1), np.random.rand(2, 2, 2)],
            [None, "CYX"],
            [None, ["A", "B"]],
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        (
            [da.random.random((1, 1, 1)), da.random.random((2, 2, 2))],
            [None, "CYX"],
            [None, ["A", "B"]],
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1)),
                xr.DataArray(np.random.rand(2, 2, 2)),
            ],
            [None, "CYX"],
            [None, ["A", "B"]],
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        (
            [
                xr.DataArray(da.random.random((1, 1, 1))),
                xr.DataArray(da.random.random((2, 2, 2))),
            ],
            [None, "CYX"],
            [None, ["A", "B"]],
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "ZYX",
            None,
        ),
        # Test many scene, different dim_order, different channel_names, second scene
        (
            [np.random.rand(1, 1, 1), np.random.rand(2, 2, 2)],
            [None, "CYX"],
            [None, ["A", "B"]],
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["A", "B"],
        ),
        (
            [da.random.random((1, 1, 1)), da.random.random((2, 2, 2))],
            [None, "CYX"],
            [None, ["A", "B"]],
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["A", "B"],
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1)),
                xr.DataArray(np.random.rand(2, 2, 2)),
            ],
            [None, "CYX"],
            [None, ["A", "B"]],
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["A", "B"],
        ),
        (
            [
                xr.DataArray(da.random.random((1, 1, 1))),
                xr.DataArray(da.random.random((2, 2, 2))),
            ],
            [None, "CYX"],
            [None, ["A", "B"]],
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["A", "B"],
        ),
        # Test filled in xarray(s)
        # no metadata should be overwritten
        (
            xr.DataArray(np.random.rand(1, 1, 1), dims=list("TYX")),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "TYX",
            None,
        ),
        (
            xr.DataArray(
                np.random.rand(1, 1, 1), dims=list("CYX"), coords={"C": ["A"]}
            ),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "CYX",
            ["A"],
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1), dims=list("TYX")),
                xr.DataArray(
                    np.random.rand(2, 2, 2), dims=list("CYX"), coords={"C": ["A", "B"]}
                ),
            ],
            None,
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1),
            "TYX",
            None,
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1), dims=list("TYX")),
                xr.DataArray(
                    np.random.rand(2, 2, 2), dims=list("CYX"), coords={"C": ["A", "B"]}
                ),
            ],
            None,
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 2, 2),
            "CYX",
            ["A", "B"],
        ),
        # Test non-standard dimensions
        (
            np.random.rand(1, 1, 1),
            "ABD",
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "ABD",
            None,
        ),
        (
            da.random.random((1, 1, 1)),
            "ABD",
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "ABD",
            None,
        ),
        (
            xr.DataArray(np.random.rand(1, 1, 1)),
            "ABD",
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "ABD",
            None,
        ),
        (
            xr.DataArray(da.random.random((1, 1, 1))),
            "ABD",
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1),
            "ABD",
            None,
        ),
        # Test that we can support many dimensions if dim_order is provided
        (
            np.random.rand(1, 2, 3, 4, 5, 6, 7, 8),
            "ABCDEFGH",
            None,
            "Image:0",
            ("Image:0",),
            (1, 2, 3, 4, 5, 6, 7, 8),
            "ABCDEFGH",
            ["Channel:0:0", "Channel:0:1", "Channel:0:2"],
        ),
        (
            [
                np.random.rand(1, 2, 3, 4, 5, 6, 7, 8),
                da.random.random((1, 2, 3, 4, 5, 6, 7, 8)),
            ],
            "ABCDEFGH",
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 2, 3, 4, 5, 6, 7, 8),
            "ABCDEFGH",
            ["Channel:0:0", "Channel:0:1", "Channel:0:2"],
        ),
        (
            [
                np.random.rand(1, 2, 3, 4, 5, 6, 7, 8),
                da.random.random((2, 3, 4, 5, 6, 7, 8, 9)),
            ],
            "ABCDEFGH",
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (2, 3, 4, 5, 6, 7, 8, 9),
            "ABCDEFGH",
            ["Channel:1:0", "Channel:1:1", "Channel:1:2", "Channel:1:3"],
        ),
        # Test that we allow metadata kwargs to be passed as lists
        # even if they "expand" out to a single scene
        (
            np.random.rand(1, 1),
            ["AX"],
            None,
            "Image:0",
            ("Image:0",),
            (1, 1),
            "AX",
            None,
        ),
        (
            np.random.rand(1, 1, 1, 1),
            None,
            [["A"]],
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1),
            "CZYX",
            ["A"],
        ),
        # Test supporting six dimensions without explicit dim_order supplied
        (
            np.random.rand(1, 2, 3, 4, 5, 6),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 2, 3, 4, 5, 6),
            "TCZYXS",
            ["Channel:0:0", "Channel:0:1"],
        ),
        # Test mismatching mapping of arrays to dim_order
        pytest.param(
            [np.random.rand(1, 1)],
            ["YX", "BAD"],
            None,
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=exceptions.ConflictingArgumentsError),
        ),
        pytest.param(
            [np.random.rand(1, 1), np.random.rand(1, 1, 1, 1)],
            ["YX", "STILL", "BAD"],
            None,
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=exceptions.ConflictingArgumentsError),
        ),
        # Test mismatching mapping of arrays to channel_names
        pytest.param(
            [np.random.rand(1, 1, 1, 1)],
            None,
            [["A"], ["B"]],
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=exceptions.ConflictingArgumentsError),
        ),
        pytest.param(
            [np.random.rand(1, 1, 1, 1), np.random.rand(1, 1, 1, 1, 1)],
            None,
            [["A"], ["B"], ["C"]],
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=exceptions.ConflictingArgumentsError),
        ),
        # Test dim string compare to dims shape
        pytest.param(
            np.random.rand(1, 1),
            "BAD",
            None,
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        # Test channel names length to size of channel dim
        pytest.param(
            np.random.rand(1, 1, 1, 1),
            None,
            ["B", "A", "D"],
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        # Test channel names provided when no channel dim present
        pytest.param(
            np.random.rand(1, 1, 1),
            None,
            ["B", "A", "D"],
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        # Test that without dim_order and with more than six dims, it raises an error
        # Our guess dim order only support up to six dims
        pytest.param(
            np.random.rand(1, 2, 3, 4, 5, 6, 7),
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=exceptions.InvalidDimensionOrderingError),
        ),
        pytest.param(
            "hello world",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=exceptions.UnsupportedFileFormatError),
        ),
    ],
)
def test_arraylike_reader(
    image: Union[types.MetaArrayLike, List[types.MetaArrayLike]],
    dim_order: Optional[Union[str, List[str]]],
    channel_names: Optional[Union[List[str], List[List[str]]]],
    set_scene: str,
    expected_scenes: Tuple[str, ...],
    expected_shape: Tuple[int, ...],
    expected_dims: str,
    expected_channel_names: Optional[List[str]],
) -> None:
    # Init
    image_container = ArrayLikeReader(
        image, dim_order=dim_order, channel_names=channel_names
    )

    test_utilities.run_image_container_checks(
        image_container=image_container,
        set_scene=set_scene,
        expected_scenes=expected_scenes,
        expected_current_scene=set_scene,
        expected_shape=expected_shape,
        expected_dtype=np.dtype(np.float64),
        expected_dims_order=expected_dims,
        expected_channel_names=expected_channel_names,
        expected_physical_pixel_sizes=None,
        # we allow both None and Dict because the user can pass an already initialized
        # xarray DataArray which has metadata as a dict
        expected_metadata_type=(
            type(None),
            dict,
        ),
    )


@pytest.mark.parametrize(
    "image, "
    "dim_order, "
    "channel_names, "
    "set_scene, "
    "expected_scenes, "
    "expected_shape, "
    "expected_dims, "
    "expected_channel_names",
    [
        # Check no metadata provided 2D
        # these are really simple just None checks
        (
            np.random.rand(1, 1),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:0:0"],
        ),
        (
            xr.DataArray(np.random.rand(1, 1)),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:0:0"],
        ),
        # Check no metadata provided 4D
        # these check that channel names are created for all
        # and specifically for xr that channel names are overwritten
        (
            np.random.rand(1, 1, 1, 1),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:0:0"],
        ),
        (
            xr.DataArray(np.random.rand(1, 1, 1, 1)),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:0:0"],
        ),
        # Test many scene, same dim_order, second scene
        (
            [np.random.rand(1, 1, 1), np.random.rand(2, 2, 2)],
            "CYX",
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (1, 2, 1, 2, 2),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:1:0", "Channel:1:1"],
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1)),
                xr.DataArray(np.random.rand(2, 2, 2)),
            ],
            "CYX",
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (1, 2, 1, 2, 2),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:1:0", "Channel:1:1"],
        ),
        # Test many scene, different dim_order, different channel_names, second scene
        (
            [np.random.rand(1, 1, 1), np.random.rand(2, 2, 2)],
            [None, "CYX"],
            [None, ["A", "B"]],
            "Image:1",
            ("Image:0", "Image:1"),
            (1, 2, 1, 2, 2),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["A", "B"],
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1)),
                xr.DataArray(np.random.rand(2, 2, 2)),
            ],
            [None, "CYX"],
            [None, ["A", "B"]],
            "Image:1",
            ("Image:0", "Image:1"),
            (1, 2, 1, 2, 2),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["A", "B"],
        ),
        # Test filled in xarray(s)
        # no metadata should be overwritten
        (
            xr.DataArray(np.random.rand(1, 1, 1), dims=list("TYX")),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:0:0"],
        ),
        (
            xr.DataArray(
                np.random.rand(1, 1, 1), dims=list("CYX"), coords={"C": ["A"]}
            ),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["A"],
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1), dims=list("TYX")),
                xr.DataArray(
                    np.random.rand(2, 2, 2), dims=list("CYX"), coords={"C": ["A", "B"]}
                ),
            ],
            None,
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 1, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:0:0"],
        ),
        (
            [
                xr.DataArray(np.random.rand(1, 1, 1), dims=list("TYX")),
                xr.DataArray(
                    np.random.rand(2, 2, 2), dims=list("CYX"), coords={"C": ["A", "B"]}
                ),
            ],
            None,
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (1, 2, 1, 2, 2),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["A", "B"],
        ),
        # Test non-standard dimensions
        (
            np.random.rand(2, 2, 2),
            "ABD",
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:0:0"],
        ),
        (
            xr.DataArray(np.random.rand(2, 2, 2)),
            "ABD",
            None,
            "Image:0",
            ("Image:0",),
            (1, 1, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:0:0"],
        ),
        # Test that we can support many dimensions if dims is provided
        (
            np.random.rand(1, 2, 3, 4, 5, 6, 7, 8),
            "ABCDEFGH",
            None,
            "Image:0",
            ("Image:0",),
            (1, 3, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:0:0", "Channel:0:1", "Channel:0:2"],
        ),
        (
            [
                np.random.rand(1, 2, 3, 4, 5, 6, 7, 8),
                da.random.random((1, 2, 3, 4, 5, 6, 7, 8)),
            ],
            "ABCDEFGH",
            None,
            "Image:0",
            ("Image:0", "Image:1"),
            (1, 3, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:0:0", "Channel:0:1", "Channel:0:2"],
        ),
        (
            [
                np.random.rand(1, 2, 3, 4, 5, 6, 7, 8),
                da.random.random((2, 3, 4, 5, 6, 7, 8, 9)),
            ],
            "ABCDEFGH",
            None,
            "Image:1",
            ("Image:0", "Image:1"),
            (1, 4, 1, 1, 1),
            dimensions.DEFAULT_DIMENSION_ORDER,
            ["Channel:1:0", "Channel:1:1", "Channel:1:2", "Channel:1:3"],
        ),
        # Test supporting six dimensions without explicit dim_order supplied
        (
            np.random.rand(1, 2, 3, 4, 5, 6),
            None,
            None,
            "Image:0",
            ("Image:0",),
            (1, 2, 3, 4, 5, 6),
            "TCZYXS",
            ["Channel:0:0", "Channel:0:1"],
        ),
        # Test that with more than 6 dims, it raises an error
        # Our guess dim order only support up to six dims
        pytest.param(
            np.random.rand(1, 2, 3, 4, 5, 6, 7),
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=exceptions.InvalidDimensionOrderingError),
        ),
    ],
)
def test_bioimage_from_array(
    image: Union[types.MetaArrayLike, List[types.MetaArrayLike]],
    dim_order: Optional[str],
    channel_names: Optional[List[str]],
    set_scene: str,
    expected_scenes: Tuple[str, ...],
    expected_shape: Tuple[int, ...],
    expected_dims: str,
    expected_channel_names: List[str],
) -> None:
    # Init
    image_container = BioImage(image, dim_order=dim_order, channel_names=channel_names)

    test_utilities.run_image_container_checks(
        image_container=image_container,
        set_scene=set_scene,
        expected_scenes=expected_scenes,
        expected_current_scene=set_scene,
        expected_shape=expected_shape,
        expected_dtype=np.dtype(np.float64),
        expected_dims_order=expected_dims,
        expected_channel_names=expected_channel_names,
        expected_physical_pixel_sizes=None,
        # we allow both None and Dict because the user can pass an already initialized
        # xarray DataArray which has metadata as a dict
        expected_metadata_type=(
            type(None),
            dict,
        ),
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, Union, get_args

import bioio_base as biob
import dask.array as da
import numpy as np
import xarray as xr
from bioio_base.standard_metadata import StandardMetadata
from bioio_base.types import MetaArrayLike
from ome_types import OME

from .ome_utils import generate_ome_channel_id
from .plugins import PluginEntry, check_type, get_array_like_plugin, get_plugins

###############################################################################

log = logging.getLogger(__name__)

###############################################################################


class BioImage(biob.image_container.ImageContainer):
    """
    BioImage takes microscopy image data types (files or arrays) of varying
    dimensions ("ZYX", "TCZYX", "CYX") and reads them as consistent 5D "TCZYX"
    ("Time-Channel-Z-Y-X") ordered array(s). The data and metadata are lazy
    loaded and can be accessed as needed.

    Parameters
    ----------
    image: biob.types.ImageLike
        A string, Path, fsspec supported URI, or arraylike to read.
    reader: Optional[Type[Reader]]
        The Reader class to specifically use for reading the provided image.
        Default: None (find matching reader)
    reconstruct_mosaic: bool
        Boolean for setting that data for this object to the reconstructed / stitched
        mosaic image.
        Default: True (reconstruct the mosaic image from tiles)
        Notes: If True and image is a mosaic, data will be fully reconstructed and
        stitched array.
        If True and base reader doesn't support tile stitching, data won't be stitched
        and instead will have an `M` dimension for tiles.
        If False and image is a mosaic, data won't be stitched and instead will have an
        `M` dimension for tiles.
        If image is not a mosaic, data won't be stitched or have an `M` dimension for
        tiles.
    use_plugin_cache: bool default False
        Boolean for setting whether to use a plugin of the installed caches rather than
        checking for installed plugins on each `BioImage` instance init.
        If True, will use the cache of installed plugins discovered last `BioImage`
        init.
    fs_kwargs: Dict[str, Any]
        Any specific keyword arguments to pass down to the fsspec created filesystem.
        Default: {}
    kwargs: Any
        Extra keyword arguments that will be passed down to the reader subclass.

    Examples
    --------
    Initialize an image then read the file and return specified slices as a numpy
    array.

    >>> img = BioImage("my_file.tiff")
    ... zstack_t8 = img.get_image_data("ZYX", T=8, C=0)

    Initialize an image, construct a delayed dask array for certain slices, then
    read only the specified chunk of data.

    >>> img = BioImage("my_file.czi")
    ... zstack_t8 = img.get_image_dask_data("ZYX", T=8, C=0)
    ... zstack_t8_data = zstack_t8.compute()

    Initialize an image with a dask or numpy array.

    >>> data = np.random.rand(100, 100)
    ... img = BioImage(data)

    Initialize an image from S3 with s3fs.

    >>> img = BioImage("s3://my_bucket/my_file.tiff")

    Initialize an image and pass arguments to the reader using kwargs.

    >>> img = BioImage("my_file.czi", chunk_dims=["T", "Y", "X"])

    Initialize an image, change scene, read data to numpy.

    >>> img = BioImage("my_many_scene.czi")
    ... img.set_scene("Image:3")
    ... img.data

    Initialize an image with a specific reader. This is useful if you know the file
    type in advance or would like to skip a few of the file format checks we do
    internally. Useful when reading from remote sources to reduce network round trips.

    >>> img = BioImage("malformed_metadata.ome.tiff", reader=readers.TiffReader)

    Data for a mosaic file is returned pre-stitched (if the base reader supports it).

    >>> img = BioImage("big_mosaic.czi")
    ... img.dims  # <Dimensions [T: 40, C: 3, Z: 1, Y: 30000, X: 45000]>

    Data for mosaic file can be explicitly returned as tiles.
    This is the same data as a reconstructed mosaic except that the tiles are
    stored in their own dimension (M).

    >>> img = BioImage("big_mosaic.czi", reconstruct_mosaic=False)
    ... img.dims  # <Dimensions [M: 150, T: 40, C: 3, Z: 1, Y: 200, X: 300]>

    Data is mosaic file but reader doesn't support tile stitching.

    >>> img = BioImage("unsupported_mosaic.ext")
    ... img.dims  # <Dimensions [M: 100, T: 1, C: 2, Z: 1, Y: 400, X: 400]>

    Notes
    -----
    If your image is made up of mosaic tiles, data and dimension information returned
    from this object will be from the tiles already stitched together.

    If you do not want the image pre-stitched together, you can use the base reader
    by either instantiating the reader independently or using the `.reader` property.
    """

    @staticmethod
    def determine_plugin(
        image: biob.types.ImageLike,
        fs_kwargs: Dict[str, Any] = {},
        use_plugin_cache: bool = False,
        **kwargs: Any,
    ) -> PluginEntry:
        """
        Determine the appropriate plugin to read a given image.

        This function identifies the most suitable plugin to read the provided image
        based on its type or file extension. It leverages the installed plugins for
        `bioio`, each of which supports a subset of image formats. If a suitable
        plugin is found, it is returned; otherwise, an error is raised.

        Parameters
        ----------
        image : biob.types.ImageLike
            The image to be read. This can be a file path (str or Path) or
            an array-like object (e.g., numpy array).
        fs_kwargs : Dict[str, Any], optional
            Additional keyword arguments to be passed to the file system handler.
        use_plugin_cache : bool, optional
            Whether to use a cached version of the plugin mapping, by default False.
        **kwargs : Any
            Additional keyword arguments for plugin-specific configurations.

        Returns
        -------
        PluginEntry
            A `PluginEntry` NamedTuple which is a wrapper of release information and
            reader references for an individual plugin.

        Raises
        ------
        exceptions.UnsupportedFileFormatError
            Raised if no suitable reader plugin can be found for the provided image.

        Notes
        -----
        This function performs the following steps:
        1. Fetches an updated mapping of available plugins,
           optionally using a cached version.
        2. If the `image` is a file path (str or Path), it checks for a matching
           plugin based on the file extension.
        3. If the `image` is an array-like object, it attempts to use the
           built-in `ArrayLikeReader`.
        4. If no suitable plugin is found, raises an `UnsupportedFileFormatError`.

        Examples
        --------
        To determine the appropriate plugin for a given image file:

        >>> image_path = "example_image.tif"
        >>> plugin = determine_plugin(image_path)
        >>> print(plugin)

        To determine the appropriate plugin for an array-like image:

        >>> import numpy as np
        >>> image_array = np.random.random((5, 5, 5))
        >>> plugin = determine_plugin(image_array)
        >>> print(plugin)

        Implementation Details
        ----------------------
        - The function first converts the image to a string representation.
        - If the image is a file path, it verifies the path and checks the file
          extension against the known plugins.
        - For each matching plugin, it tries to instantiate a reader and checks
          if it supports the image.
        - If the image is array-like, it uses a built-in reader designed for
          such objects.
        - Detailed logging is provided for troubleshooting purposes.
        """
        # Fetch updated mapping of plugins
        plugins_by_ext = get_plugins(use_cache=use_plugin_cache)

        # Try reader detection based off of file path extension
        image_str = str(type(image))
        if isinstance(image, (str, Path)):
            image_str = str(image)
            # we do not enforce_exists because it's possible
            # the path is a directory on a HTTP filesystem
            # which is impossible to check for existence
            _, path = biob.io.pathlike_to_fs(
                image, enforce_exists=False, fs_kwargs=fs_kwargs
            )

            # Check for extension in plugins_by_ext
            for format_ext, plugins in plugins_by_ext.items():
                if BioImage._path_has_extension(path, format_ext):
                    for plugin in plugins:
                        ReaderClass = plugin.metadata.get_reader()
                        try:
                            if ReaderClass.is_supported_image(
                                image,
                                fs_kwargs=fs_kwargs,
                            ):
                                return plugin

                        except FileNotFoundError as fe:
                            raise fe
                        except Exception as e:
                            log.warning(
                                f"Attempted file ({path}) load with "
                                f"reader: {ReaderClass} failed with error: {e}"
                            )

        # Use built-in ArrayLikeReader if type MetaArrayLike
        elif isinstance(image, get_args(MetaArrayLike) + (list,)):
            return get_array_like_plugin()

        # If we haven't hit anything yet, we likely don't support this file / object
        # with the current plugins installed
        raise biob.exceptions.UnsupportedFileFormatError(
            "BioImage",
            image_str,
            msg_extra=(
                "You may need to install an extra format dependency. "
                "See our list of known plugins in the bioio README here: "
                "https://github.com/bioio-devs/bioio for a list of known plugins. "
                "You can also call the 'bioio.plugins.dump_plugins()' method to "
                "report information about currently installed plugins or the "
                "'bioio.plugin_feasibility_report(image)' method to check if a "
                "specific image can be handled by the available plugins."
            ),
        )

    @staticmethod
    def _path_has_extension(path: str, extension: str) -> bool:
        """
        Examples
        --------
        >>> BioImage._path_has_extension("https://example.com/file.czi", ".czi")
        True
        >>> BioImage._path_has_extension("https://example.com/file.czi?x=123", ".czi")
        True
        >>> BioImage._path_has_extension("./file.ome.zarr", ".ome.zarr")
        True
        >>> BioImage._path_has_extension("./file.ome.zarr", ".zarr")
        True
        >>> BioImage._path_has_extension("./file.ome.zarr", ".ome")
        False
        """

        def ends_with_ext(string: str) -> bool:
            return string.lower().endswith(extension)

        return ends_with_ext(path) or ends_with_ext(path.split("?")[0])

    @staticmethod
    def _get_reader(
        image: biob.types.ImageLike,
        reader: biob.reader.Reader,
        use_plugin_cache: bool,
        fs_kwargs: Dict[str, Any],
        **kwargs: Any,
    ) -> Tuple[biob.reader.Reader, Optional[PluginEntry]]:
        """
        Initializes and returns the reader (and plugin if relevant) for the provided
        image based on provided args and/or the available bioio supported plugins
        """
        if reader is not None:
            # Check specific reader image types in a situation where a specified reader
            # only supports some of the ImageLike types.
            if not check_type(image, reader):
                raise biob.exceptions.UnsupportedFileFormatError(
                    reader.__name__, str(type(image))
                )

            return reader(image, fs_kwargs=fs_kwargs, **kwargs), None

        # Determine reader class based on available plugins
        plugin = BioImage.determine_plugin(
            image, fs_kwargs=fs_kwargs, use_plugin_cache=use_plugin_cache, **kwargs
        )
        ReaderClass = plugin.metadata.get_reader()
        return ReaderClass(image, fs_kwargs=fs_kwargs, **kwargs), plugin

    def __init__(
        self,
        image: biob.types.ImageLike,
        reader: Optional[Type[biob.reader.Reader]] = None,
        reconstruct_mosaic: bool = True,
        use_plugin_cache: bool = False,
        fs_kwargs: Dict[str, Any] = {},
        **kwargs: Any,
    ):
        try:
            self._reader, self._plugin = self._get_reader(
                image,
                reader,
                use_plugin_cache,
                fs_kwargs,
                **kwargs,
            )
        except biob.exceptions.UnsupportedFileFormatError:
            # When reading from S3 if we failed trying to read
            # try reading as an anonymous user otherwise re-raise
            # the error
            if not str(image).startswith("s3://"):
                raise

            self._reader, self._plugin = self._get_reader(
                image,
                reader,
                use_plugin_cache,
                fs_kwargs | {"anon": True},
                **kwargs,
            )

        # Store delayed modifiers
        self._reconstruct_mosaic = reconstruct_mosaic

        # Lazy load data from reader and reformat to standard dimensions
        self._xarray_dask_data: Optional[xr.DataArray] = None
        self._xarray_data: Optional[xr.DataArray] = None
        self._dims: Optional[biob.dimensions.Dimensions] = None

    @property
    def reader(self) -> biob.reader.Reader:
        """
        Returns
        -------
        reader: Reader
            The object created to read the image file type.
            The intent is that if the BioImage class doesn't provide a raw enough
            interface then the base class can be used directly.
        """
        return self._reader

    @property
    def scenes(self) -> Tuple[str, ...]:
        """
        Returns
        -------
        scenes: Tuple[str, ...]
            A tuple of valid scene ids in the file.

        Notes
        -----
        Scene IDs are strings - not a range of integers.

        When iterating over scenes please use:

        >>> for id in image.scenes

        and not:

        >>> for i in range(len(image.scenes))
        """
        return self.reader.scenes

    @property
    def current_scene(self) -> str:
        """
        Returns
        -------
        scene: str
            The current operating scene.
        """
        return self.reader.current_scene

    @property
    def current_scene_index(self) -> int:
        """
        Returns
        -------
        scene_index: int
            The current operating scene index in the file.
        """
        return self.scenes.index(self.current_scene)

    def set_scene(self, scene_id: Union[str, int]) -> None:
        """
        Set the operating scene.

        Parameters
        ----------
        scene_id: Union[str, int]
            The scene id (if string) or scene index (if integer)
            to set as the operating scene.

        Raises
        ------
        IndexError
            The provided scene id or index is not found in the available scene id list.
        TypeError
            The provided value wasn't a string (scene id) or integer (scene index).
        """
        # Update current scene on the base Reader
        # This clears the base Reader's cache
        self.reader.set_scene(scene_id)

        # Reset the data stored in the BioImage object
        self._xarray_dask_data = None
        self._xarray_data = None
        self._dims = None

    def _transform_data_array_to_bioio_image_standard(
        self,
        arr: xr.DataArray,
    ) -> xr.DataArray:
        # Determine if we need to add optionally-standard dims
        if (
            biob.dimensions.DimensionNames.Samples in arr.dims
            and biob.dimensions.DimensionNames.MosaicTile in arr.dims
        ):
            return_dims = (
                biob.dimensions.DEFAULT_DIMENSION_ORDER_WITH_MOSAIC_TILES_AND_SAMPLES
            )
        elif biob.dimensions.DimensionNames.Samples in arr.dims:
            return_dims = biob.dimensions.DEFAULT_DIMENSION_ORDER_WITH_SAMPLES
        elif biob.dimensions.DimensionNames.MosaicTile in arr.dims:
            return_dims = biob.dimensions.DEFAULT_DIMENSION_ORDER_WITH_MOSAIC_TILES
        else:
            return_dims = biob.dimensions.DEFAULT_DIMENSION_ORDER

        # Pull the data with the appropriate dimensions
        data = biob.transforms.reshape_data(
            data=arr.data,
            given_dims="".join(arr.dims),
            return_dims=return_dims,
        )

        # Pull coordinate planes
        coords: Dict[str, Any] = {}
        for d in return_dims:
            if d in arr.coords:
                coords[d] = arr.coords[d]

        # Add channel coordinate plane because it is required in BioImage
        if biob.dimensions.DimensionNames.Channel not in coords:
            coords[biob.dimensions.DimensionNames.Channel] = [
                generate_ome_channel_id(
                    image_id=self.current_scene,
                    channel_id=0,
                )
            ]

        return xr.DataArray(
            data,
            dims=tuple([d for d in return_dims]),
            coords=coords,
            attrs=arr.attrs,
        )

    @property
    def resolution_levels(self) -> Tuple[int, ...]:
        """
        Returns
        -------
        resolution_levels: Tuple[int, ...]
            A tuple of valid resolution levels in the file.
        """
        return self.reader.resolution_levels

    @property
    def current_resolution_level(self) -> int:
        """
        Returns
        -------
        current_resolution_level: int
            The currently selected resolution level.
        """
        return self.reader.current_resolution_level

    def set_resolution_level(self, resolution_level: int) -> None:
        """
        Set the operating resolution level.

        Parameters
        ----------
        resolution_level: int
            The selected resolution level to set as current.

        Raises
        ------
        IndexError
            The provided level is not found in the available list of resolution levels.
        """
        if resolution_level not in self.resolution_levels:
            raise IndexError(
                f"Resolution level {resolution_level} not found in available "
                f"resolution levels {self.resolution_levels}"
            )
        if resolution_level == self.current_resolution_level:
            return

        self.reader.set_resolution_level(resolution_level)
        # Reset the data stored in the BioImage object
        self._xarray_dask_data = None
        self._xarray_data = None
        self._dims = None

    @property
    def resolution_level_dims(self) -> Dict[int, Tuple[int, ...]]:
        """
        Returns
        -------
        resolution_level_dims: Dict[int, Tuple[int, ...]]
            resolution level dictionary of shapes.
        """
        return self.reader.resolution_level_dims

    @property
    def xarray_dask_data(self) -> xr.DataArray:
        """
        Returns
        -------
        xarray_dask_data: xr.DataArray
            The delayed image and metadata as an annotated data array.

        Notes
        -----
        If the image contains mosaic tiles, data is returned already stitched together.
        """
        if self._xarray_dask_data is None:
            if (
                # Does the user want to get stitched mosaic
                self._reconstruct_mosaic
                # Does the data have a tile dim
                and biob.dimensions.DimensionNames.MosaicTile in self.reader.dims.order
            ):
                try:
                    self._xarray_dask_data = (
                        self._transform_data_array_to_bioio_image_standard(
                            self.reader.mosaic_xarray_dask_data
                        )
                    )

                # Catch reader does not support tile stitching
                except NotImplementedError:
                    self._xarray_dask_data = (
                        self._transform_data_array_to_bioio_image_standard(
                            self.reader.xarray_dask_data
                        )
                    )

            else:
                self._xarray_dask_data = (
                    self._transform_data_array_to_bioio_image_standard(
                        self.reader.xarray_dask_data
                    )
                )

        return self._xarray_dask_data

    @property
    def xarray_data(self) -> xr.DataArray:
        """
        Returns
        -------
        xarray_data: xr.DataArray
            The fully read image and metadata as an annotated data array.

        Notes
        -----
        If the image contains mosaic tiles, data is returned already stitched together.
        Recommended to use `xarray_dask_data` for large mosaic images.
        """
        if self._xarray_data is None:
            if (
                # Does the user want to get stitched mosaic
                self._reconstruct_mosaic
                # Does the data have a tile dim
                and biob.dimensions.DimensionNames.MosaicTile in self.reader.dims.order
            ):
                try:
                    self._xarray_data = (
                        self._transform_data_array_to_bioio_image_standard(
                            self.reader.mosaic_xarray_data
                        )
                    )

                # Catch reader does not support tile stitching
                except NotImplementedError:
                    self._xarray_data = (
                        self._transform_data_array_to_bioio_image_standard(
                            self.reader.xarray_data
                        )
                    )

            else:
                self._xarray_data = self._transform_data_array_to_bioio_image_standard(
                    self.reader.xarray_data
                )

            # Remake the delayed xarray dataarray object using a rechunked dask array
            # from the just retrieved in-memory xarray dataarray
            self._xarray_dask_data = xr.DataArray(
                da.from_array(self._xarray_data.data),
                dims=self._xarray_data.dims,
                coords=self._xarray_data.coords,
                attrs=self._xarray_data.attrs,
            )

        return self._xarray_data

    @property
    def dask_data(self) -> da.Array:
        """
        Returns
        -------
        dask_data: da.Array
            The image as a dask array with standard dimension ordering.

        Notes
        -----
        If the image contains mosaic tiles, data is returned already stitched together.
        """
        return self.xarray_dask_data.data

    @property
    def data(self) -> np.ndarray:
        """
        Returns
        -------
        data: np.ndarray
            The image as a numpy array with standard dimension ordering.

        Notes
        -----
        If the image contains mosaic tiles, data is returned already stitched together.
        Recommended to use `dask_data` for large mosaic images.
        """
        return self.xarray_data.data

    @property
    def dtype(self) -> np.dtype:
        """
        Returns
        -------
        dtype: np.dtype
            Data-type of the image array's elements.
        """
        return self.xarray_dask_data.dtype

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        Returns
        -------
        shape: Tuple[int, ...]
            Tuple of the image array's dimensions.
        """
        return self.xarray_dask_data.shape

    @property
    def dims(self) -> biob.dimensions.Dimensions:
        """
        Returns
        -------
        dims: dimensions.Dimensions
            Object with the paired dimension names and their sizes.
        """
        if self._dims is None:
            self._dims = biob.dimensions.Dimensions(
                dims=self.xarray_dask_data.dims, shape=self.shape
            )

        return self._dims

    def get_image_dask_data(
        self, dimension_order_out: Optional[str] = None, **kwargs: Any
    ) -> da.Array:
        """
        Get specific dimension image data out of an image as a dask array.

        Parameters
        ----------
        dimension_order_out: Optional[str]
            A string containing the dimension ordering desired for the returned ndarray.
            Default: dimensions.DEFAULT_DIMENSION_ORDER (with or without Samples)

        kwargs: Any
            * C=1: specifies Channel 1
            * T=3: specifies the fourth index in T
            * D=n: D is Dimension letter and n is the index desired. D should not be
              present in the dimension_order_out.
            * D=[a, b, c]: D is Dimension letter and a, b, c is the list of indices
              desired. D should be present in the dimension_order_out.
            * D=(a, b, c): D is Dimension letter and a, b, c is the tuple of indices
              desired. D should be present in the dimension_order_out.
            * D=range(...): D is Dimension letter and range is the standard Python
              range function. D should be present in the dimension_order_out.
            * D=slice(...): D is Dimension letter and slice is the standard Python
              slice function. D should be present in the dimension_order_out.

        Returns
        -------
        data: da.Array
            The image data with the specified dimension ordering.

        Examples
        --------
        Specific index selection

        >>> img = BioImage("s_1_t_1_c_10_z_20.ome.tiff")
        ... c1 = img.get_image_dask_data("ZYX", C=1)

        List of index selection

        >>> img = BioImage("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_and_second = img.get_image_dask_data("CZYX", C=[0, 1])

        Tuple of index selection

        >>> img = BioImage("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_and_last = img.get_image_dask_data("CZYX", C=(0, -1))

        Range of index selection

        >>> img = BioImage("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_three = img.get_image_dask_data("CZYX", C=range(3))

        Slice selection

        >>> img = BioImage("s_1_t_1_c_10_z_20.ome.tiff")
        ... every_other = img.get_image_dask_data("CZYX", C=slice(0, -1, 2))

        Notes
        -----
        If a requested dimension is not present in the data the dimension is
        added with a depth of 1.

        See `bioio_base.transforms.reshape_data` for more details.
        """
        # If no out orientation, simply return current data as dask array
        if dimension_order_out is None and not kwargs:
            return self.dask_data

        # Transform and return
        return biob.transforms.reshape_data(
            data=self.dask_data,
            given_dims=self.dims.order,
            return_dims=dimension_order_out,
            **kwargs,
        )

    def get_image_data(
        self, dimension_order_out: Optional[str] = None, **kwargs: Any
    ) -> np.ndarray:
        """
        Read the image as a numpy array then return specific dimension image data.

        Parameters
        ----------
        dimension_order_out: Optional[str]
            A string containing the dimension ordering desired for the returned ndarray.
            Default: dimensions.DEFAULT_DIMENSION_ORDER (with or without Samples)

        kwargs: Any
            * C=1: specifies Channel 1
            * T=3: specifies the fourth index in T
            * D=n: D is Dimension letter and n is the index desired. D should not be
              present in the dimension_order_out.
            * D=[a, b, c]: D is Dimension letter and a, b, c is the list of indices
              desired. D should be present in the dimension_order_out.
            * D=(a, b, c): D is Dimension letter and a, b, c is the tuple of indices
              desired. D should be present in the dimension_order_out.
            * D=range(...): D is Dimension letter and range is the standard Python
              range function. D should be present in the dimension_order_out.
            * D=slice(...): D is Dimension letter and slice is the standard Python
              slice function. D should be present in the dimension_order_out.

        Returns
        -------
        data: np.ndarray
            The image data with the specified dimension ordering.

        Examples
        --------
        Specific index selection

        >>> img = BioImage("s_1_t_1_c_10_z_20.ome.tiff")
        ... c1 = img.get_image_data("ZYX", C=1)

        List of index selection

        >>> img = BioImage("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_and_second = img.get_image_data("CZYX", C=[0, 1])

        Tuple of index selection

        >>> img = BioImage("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_and_last = img.get_image_data("CZYX", C=(0, -1))

        Range of index selection

        >>> img = BioImage("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_three = img.get_image_dask_data("CZYX", C=range(3))

        Slice selection

        >>> img = BioImage("s_1_t_1_c_10_z_20.ome.tiff")
        ... every_other = img.get_image_data("CZYX", C=slice(0, -1, 2))

        Notes
        -----
        * If a requested dimension is not present in the data the dimension is
          added with a depth of 1.
        * This will preload the entire image before returning the requested data.

        See `bioio_base.transforms.reshape_data` for more details.
        """
        # If no out orientation, simply return current data as dask array
        if dimension_order_out is None and not kwargs:
            return self.data

        # Transform and return
        return biob.transforms.reshape_data(
            data=self.data,
            given_dims=self.dims.order,
            return_dims=dimension_order_out,
            **kwargs,
        )

    def get_stack(self, **kwargs: Any) -> np.ndarray:
        """
        Get all scenes stacked in to a single array.

        Returns
        -------
        stack: np.ndarray
            The fully stacked array. This can be 6+ dimensions with Scene being
            the first dimension.
        kwargs: Any
            Extra keyword arguments that will be passed down to the
            generate stack function.

        See Also
        --------
        bioio_base.transforms.generate_stack:
            Underlying function for generating various scene stacks.
        """
        return biob.transforms.generate_stack(self, "data", **kwargs)

    def get_dask_stack(self, **kwargs: Any) -> da.Array:
        """
        Get all scenes stacked in to a single array.

        Returns
        -------
        stack: da.Array
            The fully stacked array. This can be 6+ dimensions with Scene being
            the first dimension.
        kwargs: Any
            Extra keyword arguments that will be passed down to the
            generate stack function.

        See Also
        --------
        bioio_base.transforms.generate_stack:
            Underlying function for generating various scene stacks.
        """
        return biob.transforms.generate_stack(self, "dask_data", **kwargs)

    def get_xarray_stack(self, **kwargs: Any) -> xr.DataArray:
        """
        Get all scenes stacked in to a single array.

        Returns
        -------
        stack: xr.DataArray
            The fully stacked array. This can be 6+ dimensions with Scene being
            the first dimension.
        kwargs: Any
            Extra keyword arguments that will be passed down to the
            generate stack function.

        See Also
        --------
        bioio_base.transforms.generate_stack:
            Underlying function for generating various scene stacks.

        Notes
        -----
        When requesting an xarray stack, the first scene's coordinate planes
        are used for the returned xarray DataArray object coordinate planes.
        """
        return biob.transforms.generate_stack(self, "xarray_data", **kwargs)

    def get_xarray_dask_stack(self, **kwargs: Any) -> xr.DataArray:
        """
        Get all scenes stacked in to a single array.

        Returns
        -------
        stack: xr.DataArray
            The fully stacked array. This can be 6+ dimensions with Scene being
            the first dimension.
        kwargs: Any
            Extra keyword arguments that will be passed down to the
            generate stack function.

        See Also
        --------
        bioio_base.transforms.generate_stack:
            Underlying function for generating various scene stacks.

        Notes
        -----
        When requesting an xarray stack, the first scene's coordinate planes
        are used for the returned xarray DataArray object coordinate planes.
        """
        return biob.transforms.generate_stack(self, "xarray_dask_data", **kwargs)

    @property
    def metadata(self) -> Any:
        """
        Returns
        -------
        metadata: Any
            Passthrough to the base image reader metadata property.
            For more information, see the specific image format reader you are using
            for details on its metadata property.
        """
        return self.reader.metadata

    @property
    def ome_metadata(self) -> OME:
        """
        Returns
        -------
        metadata: OME
            The original metadata transformed into the OME specfication.
            This likely isn't a complete transformation but is guarenteed to
            be a valid transformation.

        Raises
        ------
        NotImplementedError
            No metadata transformer available.
        """
        return self.reader.ome_metadata

    @property
    def channel_names(self) -> List[str]:
        """
        Returns
        -------
        channel_names: List[str]
            Using available metadata, the list of strings representing channel names.
        """
        # Unlike the base readers, the BioImage guarantees a Channel dim
        return list(
            self.xarray_dask_data[biob.dimensions.DimensionNames.Channel].values
        )

    @property
    def physical_pixel_sizes(self) -> biob.types.PhysicalPixelSizes:
        """
        Returns
        -------
        sizes: PhysicalPixelSizes
            Using available metadata, the floats representing physical pixel sizes for
            dimensions Z, Y, and X.

        Notes
        -----
        We currently do not handle unit attachment to these values. Please see the file
        metadata for unit information.
        """
        return self.reader.physical_pixel_sizes

    @property
    def scale(self) -> biob.types.Scale:
        """
        Returns
        -------
        scale: Scale
            A Scale object constructed from the Reader's time_interval and
            physical_pixel_sizes.

        Notes
        -----
        * Combines temporal and spatial scaling information into a single object.
        * The channel scaling (`C`) is not derived from metadata and defaults to None.
        """
        return self.reader.scale

    @property
    def dimension_properties(self) -> biob.types.DimensionProperties:
        """
        Returns
        -------
        dimension_properties: DimensionProperties
            Per-dimension properties (value, type, unit) derived from the
            underlying reader.
        """
        return self.reader.dimension_properties

    @property
    def time_interval(self) -> biob.types.TimeInterval:
        """
        Returns
        -------
        sizes: Time Interval
            Using available metadata, this float represents the time interval for
            dimension T.

        Notes
        -----
        We currently do not handle unit attachment to these values. Please see the file
        metadata for unit information.
        """
        return self.reader.time_interval

    @property
    def standard_metadata(self) -> StandardMetadata:
        """
        Return a set of standardized metadata. The possible
        fields are predefined by the StandardMetadata dataclass.
        """
        return self.reader.standard_metadata

    def get_mosaic_tile_position(
        self, mosaic_tile_index: int, **kwargs: int
    ) -> Tuple[int, int]:
        """
        Get the absolute position of the top left point for a single mosaic tile.

        Parameters
        ----------
        mosaic_tile_index: int
            The index for the mosaic tile to retrieve position information for.
        kwargs: int
            The keywords below allow you to specify the dimensions that you wish
            to match. If you under-specify the constraints you can easily
            end up with a massive image stack.
                       Z = 1   # The Z-dimension.
                       C = 2   # The C-dimension ("channel").
                       T = 3   # The T-dimension ("time").

        Returns
        -------
        top: int
            The Y coordinate for the tile position.
        left: int
            The X coordinate for the tile position.

        Raises
        ------
        UnexpectedShapeError
            The image has no mosaic dimension available.
        """
        return self.reader.get_mosaic_tile_position(mosaic_tile_index, **kwargs)

    def get_mosaic_tile_positions(self, **kwargs: int) -> List[Tuple[int, int]]:
        """
        Get the absolute positions of the top left points for each mosaic tile
        matching the specified dimensions and current scene.

        Parameters
        ----------
        kwargs: int
            The keywords below allow you to specify the dimensions that you wish
            to match. If you under-specify the constraints you can easily
            end up with a massive image stack.
                       Z = 1   # The Z-dimension.
                       C = 2   # The C-dimension ("channel").
                       T = 3   # The T-dimension ("time").
                       M = 4   # The mosaic tile index

        Returns
        -------
        mosaic_tile_positions: List[Tuple[int, int]]
            List of the Y and X coordinate for the tile positions.

        Raises
        ------
        UnexpectedShapeError
            The image has no mosaic dimension available.
        NotImplementedError
            Unable to combine M dimension with other dimensions when finding
            tiles matching kwargs
        """
        if biob.dimensions.DimensionNames.MosaicTile in kwargs:
            # Don't support getting positions by M + another dim
            if len(kwargs) != 1:
                other_keys = {
                    key
                    for key in kwargs
                    if key != biob.dimensions.DimensionNames.MosaicTile
                }
                raise NotImplementedError(
                    "Unable to determine appropriate position using mosaic tile "
                    + "index (M) combined with other dimensions "
                    + f"(including {other_keys})"
                )

            return [
                self.get_mosaic_tile_position(
                    kwargs[biob.dimensions.DimensionNames.MosaicTile]
                )
            ]

        return self.reader.get_mosaic_tile_positions(**kwargs)

    @property
    def mosaic_tile_dims(self) -> Optional[biob.dimensions.Dimensions]:
        """
        Returns
        -------
        tile_dims: Optional[Dimensions]
            The dimensions for each tile in the mosaic image.
            If the image is not a mosaic image, returns None.
        """
        return self.reader.mosaic_tile_dims

    def save(
        self,
        uri: biob.types.PathLike,
        select_scenes: Optional[Union[List[str], Tuple[str, ...]]] = None,
    ) -> None:
        """
        Saves the file data to OME-TIFF format with general naive best practices.
        (Requires bioio-ome-tiff plugin installed)

        Parameters
        ----------
        uri: biob.types.PathLike
            The URI or local path for where to save the data.
            Note: Can only write to local file systems.
        select_scenes: Optional[Union[List[str], Tuple[str, ...]]]
            Which scenes in the image to save to the file.
            Default: None (save all scenes)

        Notes
        -----
        See `bioio_ome_tiff.writers.OmeTiffWriter` for more in-depth specification
        and the bioio writer registry in the README module as a whole for list of
        all available file writers.

        When reading in the produced OME-TIFF file, scenes IDs may have changed.
        This is due to how certain file and metadata formats do or do-not have IDs
        and simply names. In converting to OME-TIFF we will always store the scene
        ids in each Image's name attribute but IDs will be generated. The order of the
        scenes will be the same (or whatever order was specified / provided).
        """
        import bioio.writers as _writers  # type: ignore[attr-defined]

        OmeTiffWriter = getattr(_writers, "OmeTiffWriter", None)

        if OmeTiffWriter is None:
            log.warning(
                "TIFF writing support (OmeTiffWriter) is not available. "
                "To enable it, install the plugin:\n"
                "    pip install bioio-ome-tiff"
            )
            return

        # Get all parameters as dict of lists, or static because of unchanging values
        datas: List[biob.types.ArrayLike] = []
        dim_orders: List[Optional[str]] = []
        channel_names: List[Optional[List[str]]] = []
        image_names: List[Optional[str]] = []
        physical_pixel_sizes: List[biob.types.PhysicalPixelSizes] = []

        # Get selected scenes / handle None scenes
        if select_scenes is None:
            select_scenes = self.scenes

        # Iter through scenes to get data
        for scene in select_scenes:
            self.set_scene(scene)

            # Append this scene details
            datas.append(self.dask_data)
            dim_orders.append(self.dims.order)
            channel_names.append(self.channel_names)
            image_names.append(self.current_scene)
            physical_pixel_sizes.append(self.physical_pixel_sizes)

        # Save all selected scenes
        OmeTiffWriter.save(
            data=datas,
            uri=uri,
            dim_order=dim_orders,
            channel_names=channel_names,
            image_name=image_names,
            physical_pixel_sizes=physical_pixel_sizes,
        )

    def __str__(self) -> str:
        if self._plugin is not None:
            return (
                f"<BioImage ["
                f"plugin: {self._plugin.entrypoint.name} installed "
                f"at {datetime.datetime.fromtimestamp(self._plugin.timestamp)}, "
                f"Image-is-in-Memory: {self._xarray_data is not None}"
                f"]>"
            )

        return f"<BioImage [Image-is-in-Memory: {self._xarray_data is not None}]>"

    def __repr__(self) -> str:
        return str(self)


def _construct_img(
    image: biob.types.ImageLike, scene_id: Optional[str] = None, **kwargs: Any
) -> BioImage:
    # Construct image
    img = BioImage(image, **kwargs)

    # Select scene
    if scene_id is not None:
        img.set_scene(scene_id)

    return img


def imread_xarray_dask(
    image: biob.types.ImageLike,
    scene_id: Optional[str] = None,
    **kwargs: Any,
) -> xr.DataArray:
    """
    Read image as a delayed xarray DataArray.

    Parameters
    ----------
    image: biob.types.ImageLike
        A string, Path, fsspec supported URI, or arraylike to read.
    scene_id: Optional[str]
        An optional scene id to create the DataArray with.
        Default: None (First Scene)
    kwargs: Any
        Extra keyword arguments to be passed down to the BioImage and Reader subclass.

    Returns
    -------
    data: xr.DataArray
        The image read, scene selected, and returned as an BioIO standard shaped delayed
        xarray DataArray.
    """
    return _construct_img(image, scene_id, **kwargs).xarray_dask_data


def imread_dask(
    image: biob.types.ImageLike,
    scene_id: Optional[str] = None,
    **kwargs: Any,
) -> da.Array:
    """
    Read image as a delayed dask array.

    Parameters
    ----------
    image: biob.types.ImageLike
        A string, Path, fsspec supported URI, or arraylike to read.
    scene_id: Optional[str]
        An optional scene id to create the dask array with.
        Default: None (First Scene)
    kwargs: Any
        Extra keyword arguments to be passed down to the BioImage and Reader subclass.

    Returns
    -------
    data: da.core.Array
        The image read, scene selected, and returned as an BioIO standard shaped delayed
        dask array.
    """

    return _construct_img(image, scene_id, **kwargs).dask_data


def imread_xarray(
    image: biob.types.ImageLike,
    scene_id: Optional[str] = None,
    **kwargs: Any,
) -> xr.DataArray:
    """
    Read image as an in-memory xarray DataArray.

    Parameters
    ----------
    image: biob.types.ImageLike
        A string, Path, fsspec supported URI, or arraylike to read.
    scene_id: Optional[str]
        An optional scene id to create the DataArray with.
        Default: None (First Scene)
    kwargs: Any
        Extra keyword arguments to be passed down to the BioImage and Reader subclass.

    Returns
    -------
    data: xr.DataArray
        The image read, scene selected, and returned as an BioIO standard shaped
        in-memory DataArray.
    """
    return _construct_img(image, scene_id, **kwargs).xarray_data


def imread(
    image: biob.types.ImageLike,
    scene_id: Optional[str] = None,
    **kwargs: Any,
) -> np.ndarray:
    """
    Read image as a numpy array.

    Parameters
    ----------
    image: biob.types.ImageLike
        A string, Path, fsspec supported URI, or arraylike to read.
    scene_id: Optional[str]
        An optional scene id to create the numpy array with.
        Default: None (First Scene)
    kwargs: Any
        Extra keyword arguments to be passed down to the BioImage and Reader subclass.

    Returns
    -------
    data: np.ndarray
        The image read, scene selected, and returned as an bioio standard shaped
        np.ndarray.
    """
    return _construct_img(image, scene_id, **kwargs).data



#!/usr/bin/env python

from importlib.metadata import entry_points as _entry_points
from typing import List

# Public API list
__all__: List[str] = []

# Discover all registered writers
_eps = _entry_points(group="bioio.writers")

for ep in _eps:
    cls = ep.load()
    globals()[ep.name] = cls
    __all__.append(ep.name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Optional, Tuple

import xarray as xr
from bioio_base.dimensions import Dimensions
from bioio_base.reader import Reader as BaseReader
from fsspec.spec import AbstractFileSystem

###############################################################################


class Reader(BaseReader):
    """
    The main class of each reader plugin. This class is subclass
    of the abstract class reader (BaseReader) in bioio-base.

    Parameters
    ----------
    image: Any
        Some type of object to read and follow the Reader specification.
    fs_kwargs: Dict[str, Any]
        Any specific keyword arguments to pass down to the fsspec created filesystem.
        Default: {}

    Notes
    -----
    It is up to the implementer of the Reader to decide which types they would like to
    accept (certain readers may not support buffers for example).

    """

    _xarray_dask_data: Optional["xr.DataArray"] = None
    _xarray_data: Optional["xr.DataArray"] = None
    _mosaic_xarray_dask_data: Optional["xr.DataArray"] = None
    _mosaic_xarray_data: Optional["xr.DataArray"] = None
    _dims: Optional[Dimensions] = None
    _metadata: Optional[Any] = None
    _scenes: Optional[Tuple[str, ...]] = None
    _current_scene_index: int = 0
    # Do not provide default value because
    # they may not need to be used by your reader (i.e. input param is an array)
    _fs: "AbstractFileSystem"
    _path: str

    # Required Methods

    def __init__(image: Any, **kwargs: Any):
        raise NotImplementedError()

    @staticmethod
    def _is_supported_image(fs: "AbstractFileSystem", path: str, **kwargs: Any) -> bool:
        raise NotImplementedError()

    @property
    def scenes(self) -> Tuple[str, ...]:
        raise NotImplementedError()

    def _read_delayed(self) -> "xr.DataArray":
        raise NotImplementedError()

    def _read_immediate(self) -> "xr.DataArray":
        raise NotImplementedError()



# -*- coding: utf-8 -*-

"""Top-level package for bioio_czi."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("dummy-plugin")
except PackageNotFoundError:
    __version__ = "uninstalled"

__author__ = "bioio-devs"
__email__ = "brian.whitney@alleninstitute.org"

from .reader import Reader
from .reader_metadata import ReaderMetadata

__all__ = ["Reader", "ReaderMetadata"]
